#! /usr/bin/env python3
# _*_ coding: utf-8 _*_

#@Author:  LiuXing liuxing2@genomics.cn
#@Date:  2023-07-12 15:24:52
#@Last Modified by: LiuXing
#@Last Modified time: 2023-07-12 15:24:52



import os, sys
from optparse import OptionParser
import numpy as np
import pandas as pd
import scanpy as sc
import anndata
import seaborn as sns
import matplotlib.pyplot as plt
from scipy import sparse as sp
import logging as logg

def cn_deconvolve(adata: anndata,
                     use_rep: str = 'q05_cell_abundance_w_sf',
                     n_neighbors: int = 20,
                     resolution: float = 0.4,
                     min_dist: float = 0.2,
                     random_stat: int = 100,
                     key_added: str = None):
    """
    Cluster bins with similar cellular composition, which was calculated from cell type deconvolution result.

    Parameters
    ----------
    adata
        anndata
    use_rep
        Key value of anndata.obsm, can be used to obtain deconvolution matrix of each bin
    n_neighbors
        Number of neighbors that will be connected with the center point when constructing k-nearest neighbor graph
    resolution
        A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters. Used for leiden function.
    min_dist
        The effective minimum distance between embedded points. Smaller values will result in a more clustered/clumped embedding where 
        nearby points on the manifold are drawn closer together, while larger values will result on a more even dispersal of points. 
        The value should be set relative to the spread value, which determines the scale at which embedded points will be spread out. 
        The default of in the umap-learn package is 0.1.
    key_added
        Column name of adata.obs, which store the result of leiden clustering. 
    """
    if key_added == None:
        key_added = 'cell_neighbor'
    sc.pp.neighbors(adata, use_rep=use_rep, n_neighbors=n_neighbors, key_added='CN')
    sc.tl.leiden(adata, resolution=resolution, key_added=key_added, neighbors_key='CN')
    sc.tl.umap(adata, min_dist=min_dist, spread=1, random_state=random_stat, neighbors_key='CN')

    # analysis each Cellular Neighborhood's celltype composition according to bin level cell2location annotation matrix
    celltypes= [str[23:] for str in list(adata.obsm[use_rep])]
    # sum celltypes in each Cellular Neighborhood
    cn_pct = pd.pivot_table(adata.obs,columns=key_added,values=celltypes,aggfunc=np.sum).T
    # calculate percentage
    cn_pct = cn_pct.apply(lambda x: x/x.sum(), axis=1)
    adata.uns['CN']['cell_composition'] = cn_pct
    
def cn_cellbin(adata: anndata,
               bin_size: int = 200,
               anno: str = 'cell2loc_anno',
               n_neighbors: int = 20,
               resolution: float = 0.4,
               min_dist: float = 0.1,
               key_added: str = None,
               random_stat:int = 100,
               ):
    """
    Bin the annotated ST data in single-resolution, and cluster bins with similar cellular composition.

    Parameters
    ----------
    adata
        anndata
    bin_size
        Determines the size of bin. Take 200 as an example, calculate the cellular composition in every sparately arranged 200x200 square bin.
    anno
        Column name of anndata.obs, which contains the cell type information.
    n_neighbors
        Number of neighbors that will be connected with the center point when constructing k-nearest neighbor graph
    resolution
        A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters. Used for leiden function.
    min_dist
        The effective minimum distance between embedded points. Smaller values will result in a more clustered/clumped embedding where 
        nearby points on the manifold are drawn closer together, while larger values will result on a more even dispersal of points. 
        The value should be set relative to the spread value, which determines the scale at which embedded points will be spread out. 
        The default of in the umap-learn package is 0.1.
    key_added
        Column name of adata.obs, which store the result of leiden clustering. 
    """
    if key_added == None:
        key_added = 'cell_neighbor'
    bin_cor =[str(x[0]) + "-" + str(x[1]) for x in ((adata.obsm['spatial']//bin_size)*bin_size+(bin_size/2)).astype(int)]
    adata.obs['bin_cor'] = bin_cor
    groups = adata.obs[anno].groupby(adata.obs['bin_cor'])
    cellbin_count = pd.DataFrame(columns = adata.obs['cell2loc_anno'].unique())
    for group in groups:
        pct = group[1].value_counts()
        cellbin_count.loc[group[0]] = pct.values
    
    from sklearn.neighbors import NearestNeighbors
    #construct nearest neighbor graph
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm='ball_tree').fit(cellbin_count)
    distances, indices = nbrs.kneighbors(cellbin_count)
    X = sp.coo_matrix(([], ([], [])), shape=(indices.shape[0], 1))
    adjacency = _get_sparse_matrix_from_indices_distances_umap(indices, distances, distances.shape[0], n_neighbors)
    g = _get_igraph_from_adjacency(adjacency, directed=False)

    #cluster bins by leiden
    import leidenalg
    partition_type = leidenalg.RBConfigurationVertexPartition
    partition_kwargs = dict()
    partition_kwargs['weights'] = np.array(g.es['weight']).astype(np.float64)
    partition_kwargs['n_iterations'] = -1
    partition_kwargs['seed'] = 1
    partition_kwargs['resolution_parameter'] = resolution
    part = leidenalg.find_partition(g, partition_type, **partition_kwargs)
    groups = np.array(part.membership)
    cn_cluster = dict(zip(cellbin_count.index.values, groups))
    adata.obs[key_added] = adata.obs['bin_cor'].map(cn_cluster).astype('category')
    adata.uns['CN'] = dict()
    adata.uns['CN']['params'] = {'n_neighbors': n_neighbors,
                                 'random_state': random_stat,
                                 }
    adata.uns['CN']['leiden_cluster'] = groups
    values = cellbin_count.columns
    cellbin_count[key_added] = groups
    cn_pct = pd.pivot_table(cellbin_count, columns = key_added, values = values, aggfunc=np.sum).T
    cn_pct = cn_pct.apply(lambda x: x/x.sum()*100, axis=1)
    adata.uns['CN']['cell_composition'] = cn_pct

    #umap
    import umap
    umap_model = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist, n_components=2, random_state=random_stat)
    umap_result = umap_model.fit_transform(cellbin_count.values)
    
    adata.uns['CN']['umap'] = umap_result


def _get_sparse_matrix_from_indices_distances_umap(
    knn_indices, knn_dists, n_obs, n_neighbors
):
    rows = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    cols = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    vals = np.zeros((n_obs * n_neighbors), dtype=np.float64)

    for i in range(knn_indices.shape[0]):
        for j in range(n_neighbors):
            if knn_indices[i, j] == -1:
                continue  # We didn't get the full knn for i
            if knn_indices[i, j] == i:
                val = 0.0
            else:
                val = knn_dists[i, j]

            rows[i * n_neighbors + j] = i
            cols[i * n_neighbors + j] = knn_indices[i, j]
            vals[i * n_neighbors + j] = val

    result = sp.coo_matrix((vals, (rows, cols)), shape=(n_obs, n_obs))
    result.eliminate_zeros()
    return result.tocsr()

def _get_igraph_from_adjacency(adjacency, directed=None):
    """Get igraph graph from adjacency matrix."""
    import igraph as ig

    sources, targets = adjacency.nonzero()
    weights = adjacency[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph(directed=directed)
    g.add_vertices(adjacency.shape[0])  # this adds adjacency.shape[0] vertices
    g.add_edges(list(zip(sources, targets)))
    try:
        g.es['weight'] = weights
    except KeyError:
        pass
    if g.vcount() != adjacency.shape[0]:
        logg.warning(
            f'The constructed graph has only {g.vcount()} nodes. '
            'Your adjacency matrix contained redundant nodes.'
        )
    return g
          



import os, sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import igraph as ig
from igraph import Graph
from .scii_circos import interaction_matrix_process

def lr_link_graph_generate(interaction_matrix:pd.DataFrame,
                  cells:list,
                  separator:str="-",
                  cell_lr_separator="|",
                  reducer:int=3,
                  ) -> Graph:
    '''
    Input:
        interaction_matrix: DataFrame. Matrix contains interesting interaction in the TME region of interest, index represents cell-cell pairs
                            while column represents ligand-receptor pairs
        cells: List of cell type names, which will be used to separate cells of the cell pair.
        separator: Separator used to combine ligand with receptor genes into LR pairs.
        cell_lr_separator: Separator used to combine cells with LR genes.
        reducer: The size of a vertex = weight of the vertex/reducer
    Return:
        g: Graph containing all vertices(ligand or receptor) and their links information.
        
    '''
    #generate color palette for cells and links
    cell_palette = 'Set3'
    cell_colors = dict(zip(cells, [tuple(x[0:3]) for x in plt.get_cmap(cell_palette, len(cells)).colors]))

    #transfer matrix into igraph
    sectors, links, genes = interaction_matrix_process(interaction_matrix, cells, separator=separator, cell_lr_separator=cell_lr_separator)
    vertices = list(sectors.keys())
    vertices_color = [cell_colors[x.split(cell_lr_separator)[0]] for x in vertices]
    vertices_dict = dict(zip(vertices, range(len(vertices))))
    edges_index = [(vertices_dict[x[0]], vertices_dict[x[1]]) for x in links]
    lr_links = set([(x[0].split(cell_lr_separator)[1], x[1].split(cell_lr_separator)[1]) for x in links])
    lr_palette = "tab20"
    link_color_palette = [tuple(x[0:3]) for x in plt.get_cmap(lr_palette, len(lr_links)).colors]
    link_colors = dict(zip(lr_links, link_color_palette))
    edges_color = [link_colors[(x[0].split(cell_lr_separator)[1], x[1].split(cell_lr_separator)[1])] for x in links]

    g = Graph(n=len(vertices), edges=edges_index, directed=True)
    g.vs['name'] = vertices
    g.vs['label'] = [x.split(cell_lr_separator, 1)[1] for x in vertices]
    g.vs['color'] = vertices_color
    g.vs['weight'] = [x/reducer if x>reducer*5 else 5 for x in sectors.values()]
    g.es['weight'] = [x[2] for x in links]
    g.es['color'] = edges_color
    return g

def grap_plot(interaction_matrix:pd.DataFrame,
              cells:list,
              separator:str='-',
              cell_lr_separator:str="|",
              layout_type:str='kk',
              save:str=None,
              **kwargs,
              ):
    '''
    Input:
        interaction_matrix: DataFrame. Matrix contains interesting interaction in the TME region of interest, index represents cell-cell pairs
                            while column represents ligand-receptor pairs
        cells: List of cell type names, which will be used to separate cells of the cell pair.
        separator: Separator used to combine ligand with receptor genes into LR pairs.
        cell_lr_separator: Separator used to combine cells with LR genes.
        layout_type: Layout style used to draw the graph. Same to the layout of igraph
        save: File name of the figure that will be saved.
        **kwargs: Dictionary containing parameters of igraph plot setting.
    Return:
        None
    '''
    
    g = lr_link_graph_generate(interaction_matrix, cells, separator=separator, cell_lr_separator=cell_lr_separator)
    cell_colors = dict(zip([x.split(cell_lr_separator)[0] for x in g.vs['name']], g.vs['color']))
    link_colors = dict(zip([f"{x.source_vertex['label']}{separator}{x.target_vertex['label']}" for x in g.es], g.es['color']))

    fig, ax = plt.subplots()
    layout = g.layout(layout_type)
    layout.rotate(0)
    #draw graph
    ig.plot(g,
            layout = layout,
            vertex_size=g.vs['weight'],
            vertex_label_angle=90,
            vertex_label_dist = 0,
            vertex_label_size = 8,
            edge_width=[0.5, 4],
            edge_curved=0.2,
            edge_arrow_size=10,
            edge_arrow_width=5,
            target=ax,
            )
    #generate legend
    cell_legend_handles = []
    for cell, color in cell_colors.items():
        handle = ax.scatter(
            [], [],
            s=100,
            facecolor=color,
            label=cell,
        )
        cell_legend_handles.append(handle)
    l1 = ax.legend(
        handles=cell_legend_handles,
        title='Cell Type',
        bbox_to_anchor=(1.0, 1.0),
        bbox_transform=ax.transAxes,
    )
    lr_legend_handles = []
    for lr, color in link_colors.items():
        handle = ax.scatter(
            [], [],
            s=100,
            facecolor=color,
            label=lr,
            marker = 's',
        )
        lr_legend_handles.append(handle)
    ax.legend(
        handles=lr_legend_handles,
        title='Ligand-Receptor',
        bbox_to_anchor=(1.0, 0.6),
        bbox_transform=ax.transAxes,
    )
    fig.gca().add_artist(l1)
    fig.set_size_inches(15, 15)
    if save != None:
        fig.savefig(save)



#! /usr/bin/env python3
# _*_ coding: utf-8 _*_

#@Author:  LiuXing liuxing2@genomics.cn
#@Date:  2023-07-11 16:04:39
#@Last Modified by: LiuXing
#@Last Modified time: 2023-07-11 16:04:39

import numpy as np
import tifffile
import cv2
import os, sys
import anndata

def _color_change(value):
    digit = list(map(str, range(10))) + list("abcdef")
    a1 = digit.index(value[1]) * 16 + digit.index(value[2])
    a2 = digit.index(value[3]) * 16 + digit.index(value[4])
    a3 = digit.index(value[5]) * 16 + digit.index(value[6])
    return (a3, a2, a1)

def _get_color_list(my_color):
    if (my_color == None):
        my_color = ['#0343df', '#f97306', '#15b01a', '#e50000', '#7e1e9c', '#653700', '#ff81c0', '#929591', 
            '#6e750e', '#00ffff', '#ff796c', '#06c2ac', '#75bbfd', '#01ff07', '#cb416b', '#bf77f6', '#ceb301', 
            '#137e6d', '#516572', '#dbb40c', '#d0fefe', '#9ffeb0', '#fdaa48', '#ffcfdc', '#ffffc2', '#ac9362', 
            '#7a9703', '#96ae8d', '#b66a50', '#411900']
    rgb_mycolor = []
    for color in my_color:
        rgb_mycolor.append(list(_color_change(color)))
    return rgb_mycolor

def mask_coloring(adata: anndata, mask_file: str, 
                  anno: str='cell2loc_anno', 
                  save: str=None, save_legend: str=None):
    """
    Paint each cell in the mask with color corresponding to the palette of cell type in anndata

    Parameters
    -----------
    adata
        anndata file with cell type annotation
    mask_file
        cell mask file path
    anno
        annotation key in adata. default=cell2loc_anno
    save
        if save was specified with a file path, the colored mask will be writen to it. The file suffix should be .png|.jpg
    save_legend
        if save_legend was specified with a file path, the corresponding legend of colored mask will be writen to it. The file suffix should be .png|.jpg

    Returns
    ----------
    Tuple <colored_cell_mask, legend>
        colored cell mask: ndarray
        legend: ndarray
    """
    maskImg = tifffile.imread(mask_file)
    if (maskImg.max() == 1):
        _, labels = cv2.connectedComponents(maskImg)
    else:
        labels = maskImg
    dst = np.nonzero(labels)
    paletteKey = f"{anno}_colors"
    my_color = list(adata.uns[paletteKey]) if paletteKey in adata.uns.keys() else None
    rgb_color = _get_color_list(my_color)
    clusterDict = dict(zip(adata.obs.index.astype(int), adata.obs[anno].cat.codes))
    img_colors = [rgb_color[int(clusterDict[x])] if x in clusterDict.keys() else [255, 255, 255] for x in labels[dst]]
    new_img = np.zeros((maskImg.shape[0], maskImg.shape[1], 3), dtype = np.int8)
    new_img[dst] = img_colors

    #get legend
    cells = adata.obs[anno].cat.categories
    width = 10000
    high = 1000
    legend = np.zeros([high*len(cells), width,3])
    legend.fill(255)
    font = cv2.FONT_HERSHEY_SIMPLEX
    fontScale = 20  
    # Line thickness of 40 px
    thickness = 50
    pointSize = 100
    for i in range(len(cells)):
        color = rgb_color[i]
        cell = cells[i]
        point = (int(width/8), int((i+0.4)*high))
        cv2.circle(legend, point, pointSize, color, pointSize*2)
        org = (int(width/5), int((i+0.6)*high))
        legend = cv2.putText(legend, cell, org, font, fontScale, (0, 0, 0), thickness, cv2.LINE_AA)
    
    if save != None:
        out_dir = os.path.dirname(save)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)
        cv2.imwrite(save, new_img)
    if save_legend != None:
        out_dir = os.path.dirname(save_legend)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        cv2.imwrite(save_legend, legend)
    return new_img, legend


#! /usr/bin/env python3
# _*_ coding: utf-8 _*_

#@Author:  LiuXing liuxing2@genomics.cn
#@Date:  2023-07-12 11:01:58
#@Last Modified by: LiuXing
#@Last Modified time: 2023-07-12 11:01:58

 

import sys, os

import scanpy as sc
import anndata
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import csv
from scipy import sparse, stats
import gzip
import tifffile
import cv2
import tifffile

class Gem_Reader():
    """
    class for transfer gene expression matrix to anndata

    Parameters
    ----------
    gem_file
        gene expression matrix (gem) file path. 
        Gem file should include at least 4 columns: geneID, x, y, MIDCount. MIDCount indicates gene expression count, and can be specified by parameter count_key
        If cell segmentation has been done, gen file is consist of 5 columns: geneID, x, y, MIDCount, label. transcripts with same label belong to the same single cell.
    tissue_mask
        tissue mask file path, should be tiff format, used to extract gene expression data under the tissue covered region
    count_key
        column name of gene expression count in gem file
    cell_label_key
        column name of cell labels in gem file with single-cell segmentation.
    """
    def __init__(self,
                 gem_file: str,
                 tissue_mask: str = None,
                 count_key: str = 'MIDCount',
                 cell_label_key: str = 'label'):
        self.gem_file = gem_file
        self.tissue_mask = tissue_mask
        self.count_key = count_key
        self.cell_label_key = cell_label_key
        self._read_gem()

    def _read_gem(self) -> pd.DataFrame:
        """
        Read gene expression matrix from the given file. And extract data under tissue covered region if tissue mask file was given
        """

        columntypes = {"geneID": 'category', 
                       "x": int, 
                       'y': int, 
                       self.count_key: int,
                       }
    
        gem = pd.read_csv(self.gem_file, sep="\t", quoting=csv.QUOTE_NONE, comment="#", dtype=columntypes)

        gem.rename(columns={self.count_key : 'counts',
                            self.cell_label_key: 'label'})
        
        #extract gene expression under tissue covered region
        if self.tissue_mask != None:
            tissue_mask = tifffile.imread(self.tissue_mask)
            maxY, maxX = tissue_mask.shape
            if maxX > gem['x'].max() or maxY > gem['y'].max():
                print ("WARMING: mask is out of bounds")
                gem = gem.loc[(gem['x'] < maxX)&(gem['y'] < maxY)]
            gem = gem.loc[tissue_mask[gem['y'], gem['x']] > 0]
        self.gem = gem

    def gem2anndata(self, bin_size=50) -> anndata:
        """
        transfer gem to anndata

        Parameters
        ----------
        bin_size
            Specify bin size with this parameter. For example, bin_sie = 50 means bining spots in the same 50X50 square.
        
        Returns
        ----------
        anndata
            contains gene expression vector and spatial coordinate of each bin.
        """
    
        half_bin_size = int(bin_size/2)
    
        self.gem['x'] = (self.gem['x']//bin_size)*bin_size + half_bin_size
        self.gem['y'] = (self.gem['y']//bin_size)*bin_size + half_bin_size 
        self.gem['cell'] = self.gem['x'].astype(str) + "-" + self.gem['y'].astype(str)

        cells = self.gem['cell'].unique()
        genes = self.gem['geneID'].unique()
    
        cells_dict = dict(zip(cells, range(0, len(cells))))
        genes_dict = dict(zip(genes, range(0, len(genes))))
        rows = self.gem['cell'].map(cells_dict)
        cols = self.gem['geneID'].map(genes_dict)
        expMtx = sparse.csr_matrix((self.gem[self.count_key].values, (rows, cols)), shape=(cells.shape[0], genes.shape[0]), dtype=np.int32)

        obs = pd.DataFrame(index = cells)
        var = pd.DataFrame(index = genes)
        adata = anndata.AnnData(X = expMtx, obs = obs, var = var)
        positions = np.array(list(map(lambda x: [int(v) for v in x.strip().split("-")], adata.obs.index)))
        adata.obs['x'] = positions[:,0]
        adata.obs['y'] = positions[:,1]
        adata.obsm['spatial'] = adata.obs[['x', 'y']].values
        return adata

    def cellbin2anndata(self) -> anndata:
        """
        transfer gem with single-cell segmentation to anndata

        Returns
        -----------
        anndata
            contains gene expression vector and sptial coordinate of each cell 
        """

        gem = self.gem[self.gem[self.cell_label_key]!=0]
    
        cells = gem[self.cell_label_key].unique()
        genes = gem['geneID'].unique()
        cells_dict = dict(zip(cells, range(0, len(cells))))
        genes_dict = dict(zip(genes, range(0, len(genes))))
        rows = gem[self.cell_label_key].map(cells_dict)
        cols = gem['geneID'].map(genes_dict)

        expMtx = sparse.csr_matrix((gem[self.count_key].values, (rows, cols)), shape=(cells.shape[0], genes.shape[0]), dtype=np.int32)

        obs = pd.DataFrame(index = cells)
        var = pd.DataFrame(index = genes)
        adata = anndata.AnnData(X = expMtx, obs = obs, var = var)
        spatialgroup = gem[['x', 'y']].groupby(gem[self.cell_label_key])
        spatialdf = spatialgroup.agg(lambda x: (x.max()+x.min())/2)
        spatialdf = spatialdf.reset_index()
        spatialdf[self.cell_label_key] = spatialdf[self.cell_label_key].astype('category').cat.reorder_categories(cells, ordered=True)
        spatialdf.sort_values(self.cell_label_key, inplace=True)
        adata.obsm['spatial'] = spatialdf[['x', 'y']].values
        return adata

    def gem_with_cellmask_2anndata(self, cell_mask: str) -> anndata:
        """
        extract single-cell gem based on the cell mask and transfer it to anndata

        Parameters
        -----------
        cell_mask
            cell mask file path, should be tiff format, used to extract gene expression of each single-cell
        
        Returns
        -----------
        anndata
            contains gene expression vector and sptial coordinate of each cell
        """

        mask = cv2.imread(cell_mask, -1)
        if (mask.max() == 1):
            _, labels = cv2.connectedComponents(mask)
        else:
            labels = mask
        tissuedf = pd.DataFrame()
        dst = np.nonzero(labels)

        tissuedf['x'] = dst[1]
        tissuedf['y'] = dst[0]
        tissuedf[self.cell_label_key] = labels[dst]

        res = pd.merge(self.gem, tissuedf, on=['x', 'y'], how='inner')
        self.gem = res
        adata = self.cellbin2anndata()
        return adata



import numpy as np
import pandas as pd
import squidpy as sq
import os, sys
from itertools import product
from scipy import sparse
import time
import tensorly as tl
from tensorly.decomposition import non_negative_tucker, non_negative_tucker_hals
from tqdm import tqdm
import logging as logg
import torch
import random

from stereosite.scii import _generate_LRpairs, _get_LR_connect_matrix, _get_LR_intensity, _permutation_test,  _result_combined, _mh_translate


format="%(asctime)s-%(levelname)s-%(message)s"
logg.basicConfig(format=format, level=logg.INFO)

def get_keys(d, value):
    return [k for k, v in d.items() if v == value]

def _sorted(s, num):
    tmp = s.sort_values(ascending=False)[:num].index
    tmp.index = range(num)
    return tmp

def find_max_column_indices(array):
    max_column_indices = np.argmax(array, axis=1)
    return max_column_indices

class InteractionTensor():
    """
    Class for constructing and analyzing Spatial Cellular Interaction Intensity (SCII) matrices
    based on spatial transcriptomics data.

    The InteractionTensor class provides methods to build SCII matrices, process them,
    evaluate tensor ranks, and perform tensor factorization for further analysis.

    Attributes:
    - adata: anndata
        The spatial transcriptomics data.
    - interactionDB: str
        Database of ligand-receptor pairs.
    """
    
    def __init__(self, adata, interactionDB: str=None):
        # Save variables for this class
        self.adata = adata
        self.interactionDB = interactionDB

def build_SCII(interactiontensor: InteractionTensor=None, radius = 200, #int or dict 
               scale:float=0.5,
               coord_type: str="generic",
               window_size: int=400, anno_col: str='cell2loc_anno', use_raw: bool=True, interactionDB: str=None) -> list:
    """
    Construct a Spatial Cellular Interaction Intensity (SCII) matrix based on spatial transcriptomics data.

    This function calculates the interactions between different cell types within specified spatial windows
    and generates a CCI matrix containing interaction strengths.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    radius : int
        The radius for spatial neighbor search, by default 200 µm. 
        If the ligand-receptor pairs have been clustered into different types, the distance_threshold can receive a dictionary with 
        LR types and corresponding distance thresholds. e.g: {'Secreted Signaling': 200, 'ECM-Receptor': 200, 'Cell-Cell Contact': 30}
    scale : float
        The distance between adjancent spots, the unit is µm. For Stereo-chip, scale=0.5. default=0.5
    coord_type : str
        The coordinate type for spatial neighbor calculation, by default "generic".
    window_size : int
        The size of the spatial windows, by default 400µm.
    anno_col : str
        Column name containing cell annotations, by default 'cell2loc_anno'.
    use_raw : bool
        whether use raw counts to build CCI matrix, by default True.
    interactionDB : str
        Database of ligand-receptor pairs.

    Returns
    -------
    list of pandas.DataFrame
        A list of DataFrames containing interaction strengths between cell types in each window.
    """
    interactionDB = interactiontensor.interactionDB
    adata = interactiontensor.adata

    interactiontensor.radius = radius
    interactiontensor.scale = scale
    interactiontensor.coord_type = coord_type
    interactiontensor.window_size = window_size
    interactiontensor.anno_col = anno_col
    interactiontensor.use_raw = use_raw

    time_start=time.time()

    if use_raw and adata.raw != None:
        adata = adata.raw.to_adata()
    #else:
    #    sc.pp.normalize_total(adata, inplace=True, target_sum=1e4)
    #    sc.pp.log1p(adata)

    # filter LR genes
    logg.info("Filter LR genes")
    filter_LRpairs_dict, adata_sub = _generate_LRpairs(interactionDB=interactionDB, adata=adata)

    # nearest neighbor graph
    logg.info("Create spatial neighbor graph")
    if isinstance(radius, int):
        radius = radius/scale
        sq.gr.spatial_neighbors(adata_sub, radius=radius, coord_type=coord_type, key_added=str(radius))
    elif isinstance(radius, dict):
        for LRtype, radii in radius.items():
            radii = radii/scale
            radius[LRtype] = radii
            sq.gr.spatial_neighbors(adata_sub, radius=radii, coord_type=coord_type, key_added=str(radii))
    else:
        raise Exception("radius type must be int or dict, but we got {0}".format(type(radius)))
 
    exp_connect_list = []
    filter_LRpairs = []
    for LRtype, LRpair_list in filter_LRpairs_dict.items():
        if isinstance(radius, int):
            connect_matrix = adata_sub.obsp[f"{radius}_connectivities"]
        elif isinstance(radius, dict):
            if not LRtype in radius.keys():
                raise Exception(f"The radius threshold of LRtype {LRtype} in the interactionDB was not specified by the radius parameter: {radius}")
            connect_matrix = adata_sub.obsp[f"{radius[LRtype]}_connectivities"]
        for genes in LRpair_list:
            filter_LRpairs.append(genes)
            exp_connect_matrix = _get_LR_connect_matrix(adata_sub, genes, connect_matrix)
            exp_connect_list.append(exp_connect_matrix)
    interactiontensor.filter_LRpair = filter_LRpairs

    interactiontensor.exp_connect_list = exp_connect_list

    window_size = window_size/scale
    adata_sub.obs[f'x_{window_size}'] = (adata_sub.obsm['spatial'][:, 0]//window_size)*window_size+(window_size/2)
    adata_sub.obs[f'y_{window_size}'] = (adata_sub.obsm['spatial'][:, 1]//window_size)*window_size+(window_size/2)
    adata_sub.obs[f'bin{window_size}'] = adata_sub.obs[f'x_{window_size}'].astype('str') + '_' + adata_sub.obs[f'y_{window_size}'].astype("str")

    adata_sub.obs['cellname'] = adata_sub.obs.index.tolist()
    df = adata_sub.obs[['cellname', f'bin{window_size}']].reset_index()
    df['index'] = df.index.tolist()
    bin_group = df[['index', f'bin{window_size}']].groupby(f'bin{window_size}')
    indices = [indices for t, indices in bin_group.groups.items()]

    interactiontensor.indices = indices

    logg.info("Start build SCII matrix")
    results = []
    # i=0
    celltype = adata_sub.obs[anno_col].values.unique().tolist()
    combinations = list(product(celltype, repeat=2))
    for i in tqdm(range(len(indices))):   
        store_mt = pd.DataFrame(np.zeros((len(combinations), len(filter_LRpairs))), index = ['_'.join(comb) for comb in combinations], columns=["-".join(x) for x in filter_LRpairs]).astype('float32')

        adata_sub_sub = adata_sub[indices[i], :]

        cell_types = adata_sub_sub.obs[anno_col].unique()
        cell_type_dict = dict(zip(cell_types, range(0, len(cell_types))))
        cellTypeIndex = adata_sub_sub.obs[anno_col].map(cell_type_dict).astype(int).values

        cellTypeNumber = len(adata_sub_sub.obs[anno_col].unique())

        for j in range(len(filter_LRpairs)):
            # j=0
            lr = "-".join(filter_LRpairs[j])

            data = exp_connect_list[j].tocsr()[indices[i], :][:, indices[i]].tocoo()

            senders = cellTypeIndex[data.row]
            receivers = cellTypeIndex[data.col]
            interaction_matrix = sparse.csr_matrix((data.data, (senders, receivers)), shape=(cellTypeNumber, cellTypeNumber))

            nonzero_row, nonzero_col = interaction_matrix.nonzero()
            for r in range(len(nonzero_row)):
                for c in range(len(nonzero_col)):
                    cellpair = get_keys(cell_type_dict, nonzero_row[r])[0] + "_" + get_keys(cell_type_dict, nonzero_col[c])[0]

                    store_mt.loc[cellpair][lr] = interaction_matrix[nonzero_row[r], nonzero_col[c]]
        results.append(store_mt)

    interactiontensor.lr_mt_list = results
    time_end=time.time()
    logg.info(f"Finish build CCI matrix - time cost {(((time_end - time_start) / 60) / 60)} h")


def process_SCII(interactiontensor: InteractionTensor=None, 
            zero_remove: bool=True,
            log_data: bool=True) -> np.ndarray:

    """
    Process the Spatial Cellular Interaction Intensity (SCII) matrix list.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    bin_zero_remove : bool
        Flag indicating whether to remove spatial bins with zero intensity, by default True.
    log_data : bool
        Flag indicating whether to log-transform the data, by default True.

    Returns
    -------
    np.ndarray
        Processed 3D Celltype-Celltype Interaction (CCI) matrix.
    """

    time_start=time.time()
    lr_mt_list = interactiontensor.lr_mt_list
    final_mt = np.dstack(lr_mt_list)

    if zero_remove == True:
        window_zero_indices = np.all(final_mt == 0, axis=(0,1))
        lrpair_zero_indices = np.all(final_mt == 0, axis=(0,2))
        cellpair_zero_indices = np.all(final_mt == 0, axis=(1,2))
        
        logg.info(f"{sum(window_zero_indices)} window, {sum(lrpair_zero_indices)} lrpair, {sum(cellpair_zero_indices)} cellpair have zero intensity")
        
        interactiontensor.zero_indices = [cellpair_zero_indices, lrpair_zero_indices, window_zero_indices]
        
        final_mt = final_mt[:, :, ~window_zero_indices]
        final_mt = final_mt[:, ~lrpair_zero_indices, :]
        final_mt = final_mt[~cellpair_zero_indices, :, :]
        
        cellpair = lr_mt_list[0].index[~cellpair_zero_indices]
        interactiontensor.cellpair = cellpair
    
        lrpair = lr_mt_list[0].columns[~lrpair_zero_indices]
        interactiontensor.lrpair = lrpair
        
        lr_mt_list_filter = []
        for i in range(len(lr_mt_list)):
            lr_mt_list_filter.append(lr_mt_list[i].loc[cellpair][lrpair])
        lr_mt_list_filter = [k for k,v in zip(lr_mt_list_filter, ~window_zero_indices) if v]
        interactiontensor.lr_mt_list_filter = lr_mt_list_filter
        
        indices = interactiontensor.indices

        indices_filter = [k for k,v in zip(indices, (~window_zero_indices).tolist()) if v]
        interactiontensor.indices_filter = indices_filter

    else:
        interactiontensor.cellpair = lr_mt_list[0].index
        interactiontensor.lrpair = lr_mt_list[0].columns
        interactiontensor.zero_indices = []
        interactiontensor.lr_mt_list_filter = lr_mt_list
        interactiontensor.indices_filter = indices

    if log_data:
        final_mt = np.log1p(final_mt)

    interactiontensor.cci_matrix = final_mt

    time_end=time.time()
    logg.info(f"Finish processing LR matrix - time cost {(((time_end - time_start) / 60) / 60)} h")

def evaluate_ranks(interactiontensor: InteractionTensor=None, 
                   num_TME_modules: int=20, num_cellpair_modules: int=20, 
                   device: str="cuda:0", init: str='svd', 
                   n_iter_max: int=10, 
                   method: str='tucker', 
                   use_gpu: bool=True,
                   backend:str='pytorch'
                   ) -> np.ndarray:
    """
    Evaluate reconstruction error of tensor decomposition of Cell-Cell Interaction (CCI) tensor matrix.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    num_TME_modules : int, optional
        Number of modules for the TME (Tumor Microenvironment) to test, by default 20.
    num_cellpair_modules : int, optional
        Number of modules for cell-cell pairs and ligand-receptor pairs to test, by default 20.
    device : str, optional
        Device for computation ('cpu' or 'cuda:X'), by default 'cuda:0'.
    init : str, optional
        Initialization method for tensor decomposition, by default 'svd'.
    n_iter_max : int, optional
        Maximum number of iterations for tensor decomposition, by default 10.
    method : str, optional
        Method for tensor decomposition ('tucker' or 'hals'), by default 'tucker'.
    use_gpu : bool, optional
        Whether to use GPU for computation, by default True.

    Returns
    -------
    np.ndarray
        A matrix representing evaluated reconstruction errors.
    """

    time_start=time.time()
    tl.set_backend(backend)
    dat = interactiontensor.cci_matrix
    torch.cuda.empty_cache()
    # pal = sns.color_palette('bright',10)
    # palg = sns.color_palette('Greys',10)
    num_TME_modules = num_TME_modules + 1
    mat1 = np.zeros((num_TME_modules,num_cellpair_modules))
    # mat2 = np.zeros((num_TME_modules,num_cellpair_modules))
    if use_gpu:
        tensor = tl.tensor(dat, device=device)
    else:
        tensor = tl.tensor(dat)
        tensor = tensor.to("cpu")

    for i in tqdm(range(2,num_cellpair_modules)):
        for j in range(1,num_TME_modules):
            # we use NNTD as described in the paper

            if method == 'tucker':
                facs_overall = non_negative_tucker(tensor,rank=[i,i, j],random_state = 2337, init=init, n_iter_max=n_iter_max)
                # facs_overall = [factor.cpu() for factor in facs_overall]
                mat1[j,i] = np.mean((dat- tl.to_numpy(tl.tucker_to_tensor((facs_overall[0],facs_overall[1]))))**2)
                # mat2[j,i] = np.linalg.norm(dat - tl.to_numpy(tl.tucker_to_tensor((facs_overall[0],facs_overall[1]))) ) / np.linalg.norm(dat)
            elif method == 'hals':
                facs_overall = non_negative_tucker_hals(tensor,rank=[i,i, j],random_state = 2337, init=init, n_iter_max=n_iter_max)
                mat1[j,i] = np.mean((dat- tl.to_numpy(tl.tucker_to_tensor((facs_overall[0],facs_overall[1]))))**2)
    interactiontensor.reconstruction_errors = mat1
    time_end=time.time()
    logg.info(f"Finish eval SCII tensor rank - time cost {(((time_end - time_start) / 60) / 60)} h")
    return mat1

def SCII_Tensor(interactiontensor: InteractionTensor=None, rank: list=[8,8,8], random_state: int=32, init: str="svd", n_iter_max: int=100, backend: str="pytorch",
                device='cuda:0'):
    """
    Perform tensor factorization and analysis on a Cell-Cell Interaction (CCI) matrix.

    This function reads CCI matrix data, performs non-negative tensor factorization, generates visualizations,
    and saves the results for further analysis.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    rank : list, optional
        Rank for the non-negative tensor factorization, by default [8, 8, 8].
    random_state : int, optional
        Random state for reproducibility, by default 32.
    init : str, optional
        Initialization method for tensor factorization, by default 'svd'.
    n_iter_max : int, optional
        Maximum number of iterations for tensor factorization, by default 100.
    backend : str, optional
        Backend for tensor operations, by default 'pytorch'.
    device : str, optional
        Device for computation ('cpu' or 'cuda:X'), by default 'cuda:0'.
    """

    time_start=time.time()

    torch.cuda.empty_cache()
    tl.set_backend(backend)

    tensor = tl.tensor(interactiontensor.cci_matrix, device=device)
    core, factors = non_negative_tucker(tensor, rank=rank, random_state = random_state, init=init, n_iter_max=n_iter_max)

    factors = [x.data.cpu().numpy() for x in factors]
    core = core.data.cpu().numpy()

    interactiontensor.factors = factors
    interactiontensor.core = core

    cellpair = interactiontensor.cellpair
    lrpair = interactiontensor.lrpair

    tme_cluster = find_max_column_indices(factors[2])
    interactiontensor.tme_cluster = tme_cluster

    indices_filter = interactiontensor.indices_filter

    adata = interactiontensor.adata
    adata.obs['TME_module'] = None
    for k,v in enumerate(tme_cluster):
        adata.obs['TME_module'].iloc[indices_filter[k]] = str(v)

    adata.obs.TME_module = adata.obs.TME_module.astype('category')
    adata.obs.TME_module = adata.obs.TME_module.cat.set_categories(np.arange(adata.obs.TME_module.cat.categories.astype('int').max()+1).astype('str'))
    interactiontensor.adata = adata

    lr_df = pd.DataFrame(factors[1], index = lrpair)
    interactiontensor.lr_factor = lr_df

    cc_df = pd.DataFrame(factors[0], index = cellpair)
    interactiontensor.cc_factor = cc_df

    time_end=time.time()
    logg.info(f"Finish SCII tensor - time cost {(((time_end - time_start) / 60) / 60)} h")


def SCII_Tensor_multiple(interactiontensor: InteractionTensor=None, rank: list=[8,8,8], random_state: int=32, init: str="svd", n_iter_max: int=100, backend: str="pytorch",
                        device: str='cuda:0'):
    """
    Perform tensor factorization and analysis on a Cell-Cell Interaction (CCI) matrix.

    This function reads CCI matrix data, performs non-negative tensor factorization, generates visualizations,
    and saves the results for further analysis.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    rank : list, optional
        Rank for the non-negative tensor factorization, by default [8, 8, 8].
    random_state : int, optional
        Random state for reproducibility, by default 32.
    init : str, optional
        Initialization method for tensor factorization, by default 'svd'.
    n_iter_max : int, optional
        Maximum number of iterations for tensor factorization, by default 100.
    backend : str, optional
        Backend for tensor operations, by default 'pytorch'.
    top_n_cc : int, optional
        Number of top cell-cell pairs to consider for visualization, by default 3.
    top_n_lr : int, optional
        Number of top LR pairs to consider for visualization, by default 3.
    figsize : tuple, optional
        Size of the figure for heatmaps, by default (8,15).
    device : str, optional
        Device for computation ('cpu' or 'cuda:X'), by default 'cuda:0'.
    """

    time_start=time.time()

    torch.cuda.empty_cache()
    tl.set_backend(backend)
    tensor = tl.tensor(interactiontensor.cci_matrix, device=device)

    core, factors = non_negative_tucker(tensor, rank=rank, random_state = random_state, init=init, n_iter_max=n_iter_max)

    factors = [x.data.cpu().numpy() for x in factors]
    core = core.data.cpu().numpy()

    interactiontensor.factors = factors
    interactiontensor.core = core

    cellpair = interactiontensor.cellpair
    lrpair = interactiontensor.lrpair

    num_TME_modules=rank[2]

    tme_cluster = find_max_column_indices(factors[2])
    interactiontensor.tme_cluster = tme_cluster
 
    patient_id = list(interactiontensor.module_lr_mt.keys())
    module_lr_mt = interactiontensor.module_lr_mt
    
    individual_patient = [[k]*v.shape[2] for k,v in zip(patient_id, module_lr_mt.values())]
    individual_patient = [item for sublist in individual_patient for item in sublist]
    interactiontensor.patient_map = individual_patient 

    individual_module = [np.arange(v.shape[2]) for v in module_lr_mt.values()]
    individual_module = [item for sublist in individual_module for item in sublist]
    interactiontensor.module_map = individual_module

    adata_list = interactiontensor.adata

    adata_list[0].obs['tmp'] = [str(random.randint(0, num_TME_modules-1)) for _ in range(adata_list[0].n_obs)]

    ids = patient_id
    for i in range(len(ids)):

        tme_cluster_idx = [v for k,v in zip(individual_patient, tme_cluster) if k == ids[i]]

        tme_module_idx = [v for k,v in zip(individual_patient, individual_module) if k == ids[i]]

        adata_list[i].obs['TME_meta_module'] = None

        adata_list[i].obs['TME_meta_module'] = adata_list[i].obs['TME_module'].map(dict(zip([str(x) for x in tme_module_idx], [str(x) for x in tme_cluster_idx])))

        adata_list[i].obs['TME_meta_module'] =  adata_list[i].obs['TME_meta_module'].astype('category')

    interactiontensor.adata_list_TME = adata_list

    lr_df = pd.DataFrame(factors[1], index = lrpair)
    interactiontensor.lr_factor = lr_df

    cc_df = pd.DataFrame(factors[0], index = cellpair)
    interactiontensor.cc_factor = cc_df

    time_end=time.time()
    logg.info(f"Finish multiple sample SCII tensor - time cost {(((time_end - time_start) / 60) / 60)} h")

def top_pair(interactiontensor: InteractionTensor=None, pair='cc', top_n: int=15):
    """
    Identify the top CellPairs or LRpairs based on their loadings.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    pair : str, optional
        Type of pair to consider, either 'cc' for CellPairs or 'lr' for LRpairs, by default 'cc'.
    top_n : int, optional
        Number of top pairs to identify, by default 15.
    """
    if pair == 'cc':
        cc_df = interactiontensor.cc_factor
        top_ccpair = cc_df.loc[cc_df.sum(axis=1).sort_values(ascending=False).index[0:top_n]]
        #interactiontensor.top_ccpair=top_ccpair
        return top_ccpair
    elif pair == 'lr':
        lr_df = interactiontensor.lr_factor
        top_lrpair = lr_df.loc[lr_df.sum(axis=1).sort_values(ascending=False).index[0:top_n]]
        #interactiontensor.top_lrpair= top_lrpair
        return top_lrpair

def interaction_select_multiple(interactiontensor: InteractionTensor=None, sample: str=None, 
                                     tme_module: int=0, cellpair_module: int=0, lrpair_module: int=0, 
                                     n_lr: int=15, n_cc: int=5, 
                                ):
    """
    Plot the mean intensity heatmap for a specific TME module, CellPair module, LRPair module, and patient sample.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    sample : str, optional
        Patient sample identifier, by default None.
    tme_module : int, optional
        Index of the TME module to plot, by default 0.
    cellpair_module : int, optional
        Index of the CellPair module to plot, by default 0.
    lrpair_module : int, optional
        Index of the LRPair module to plot, by default 0.
    n_lr : int, optional
        Number of LRPairs to include in the plot, by default 15.
    n_cc : int, optional
        Number of CellPairs to include in the plot, by default 5.
    """

    module_lr_mt = interactiontensor.module_lr_mt
    cellpair = interactiontensor.cellpair
    lrpair = interactiontensor.lrpair
    tme_cluster = interactiontensor.tme_cluster

    individual_patient = interactiontensor.patient_map
    individual_module = interactiontensor.module_map

    factor_cc = interactiontensor.cc_factor
    factor_lr = interactiontensor.lr_factor

    top_lrpair = factor_lr.apply(lambda x: _sorted(x, n_lr))
    
    sub_cci_matrix = module_lr_mt[sample][:, :, [y for x,y,z in zip(individual_patient, individual_module, tme_cluster.tolist()) if x==sample and z == tme_module]]    
    mean_sub_mt = np.mean(sub_cci_matrix, axis=2)
    mean_df = pd.DataFrame(mean_sub_mt, index = cellpair, columns = lrpair)    
    mean_mt = mean_df[top_lrpair[lrpair_module][0:n_lr]].loc[_sorted(factor_cc[cellpair_module],n_cc).tolist()]    

    return mean_mt

def interaction_select(interactiontensor: InteractionTensor=None, 
                            tme_module: int=0, cellpair_module: int=0, lrpair_module: int=0, 
                            n_lr: int=15, n_cc: int=5, 
                      ):
    """
    Plot the mean intensity heatmap for a specific TME module, CellPair module, and LRPair module.

    Parameters
    ----------
    interactiontensor : InteractionTensor
        The InteractionTensor of the spatial transcriptomics data.
    tme_module : int, optional
        Index of the TME module to plot, by default 0.
    cellpair_module : int, optional
        Index of the CellPair module to plot, by default 0.
    lrpair_module : int, optional
        Index of the LRPair module to plot, by default 0.
    n_lr : int, optional
        Number of LRPairs to include in the plot, by default 15.
    n_cc : int, optional
        Number of CellPairs to include in the plot, by default 5.
    """

    lr_mt_list = interactiontensor.lr_mt_list_filter
    cellpair = interactiontensor.cellpair
    lrpair = interactiontensor.lrpair

    tme_cluster = interactiontensor.tme_cluster

    factor_cc = interactiontensor.cc_factor
    factor_lr = interactiontensor.lr_factor

    top_lrpair = factor_lr.apply(lambda x: _sorted(x, n_lr))
    
    sub_scii_matrix = [k for k,v in zip(lr_mt_list, pd.Series(tme_cluster).isin([tme_module])) if v]
    sub_mt = np.dstack(sub_scii_matrix)
    mean_sub_mt = np.mean(sub_mt, axis=2)
    mean_df = pd.DataFrame(mean_sub_mt, index=cellpair, columns=lrpair)
    mean_mt = mean_df[top_lrpair[lrpair_module][0:n_lr]].loc[_sorted(factor_cc[cellpair_module], n_cc).tolist()]

    return mean_mt

def merge_data(interactiontensor_list: list=None, patient_id: list=None) -> InteractionTensor:
    """
    Merge data from multiple InteractionTensor.

    Parameters
    ----------
    interactiontensor_list : list, optional
        List of InteractionTensor instances to merge, by default None.
    patient_id : list
        List of patient id

    Returns
    -------
    InteractionTensor
        A new InteractionTensor instance containing merged data.
    """
    lr_mt_list = []
    lr_mt_list_filter = []
    adata_list = []
    #zero_indices_list = []
    tme_cluster_list = []
    indice_list = []
    indice_filter_list = []
    for interactiontensor in interactiontensor_list:
        lr_mt_list.append(interactiontensor.lr_mt_list)
        lr_mt_list_filter.append(interactiontensor.lr_mt_list_filter)
        adata_list.append(interactiontensor.adata)
        #zero_indices_list.append(interactiontensor.window_zero_indices)
        tme_cluster_list.append(interactiontensor.tme_cluster)
        indice_list.append(interactiontensor.indices)
        indice_filter_list.append(interactiontensor.indices_filter)

    tmp = InteractionTensor(adata_list)

    common_columns = set(lr_mt_list_filter[0][0].columns)
    for i in range(len(lr_mt_list_filter)):
        common_columns = common_columns.intersection(set(lr_mt_list_filter[i][0].columns))
    
    common_rows = set(lr_mt_list_filter[0][0].index)
    for i in range(len(lr_mt_list_filter)):
        common_rows = common_rows.intersection(set(lr_mt_list_filter[i][0].index))
        
    tmp.lrpair = list(common_columns)
    tmp.cellpair = list(common_rows)

    for i in range(len(lr_mt_list_filter)):
        for j in range(len(lr_mt_list_filter[i])):
            lr_mt_list_filter[i][j] = lr_mt_list_filter[i][j][list(common_columns)].loc[list(common_rows)]

    cci_matrix = []
    for i in range(len(lr_mt_list_filter)):
        cci_matrix.append(np.dstack(lr_mt_list_filter[i]))

    for i in range(len(cci_matrix)):
        cci_matrix[i] = np.log1p(cci_matrix[i])

    tmp.cci_matrix_individual = cci_matrix

    indice = [item for indice in indice_filter_list for item in indice]

    tmp.indices_filter = indice

    module_lr_mt = {}
    for i in range(len(cci_matrix)):
        result = np.zeros((cci_matrix[0].shape[0], cci_matrix[0].shape[1], np.unique(tme_cluster_list[i]).max()+1))

        for cluster in range(np.unique(tme_cluster_list[i]).max()+1):
            cluster_mask = (tme_cluster_list[i] == cluster)  # Create a boolean mask for the current cluster
            cluster_data = cci_matrix[i][:, :, cluster_mask]  # Extract data for the current cluster
            result[:, :, cluster] = np.mean(cluster_data, axis=2)  # Calculate the mean along the third dimension
            module_lr_mt[patient_id[i]] = result
    tmp.module_lr_mt = module_lr_mt

    final_mt = np.nan_to_num(np.dstack(list(module_lr_mt.values())))

    tmp.cci_matrix = final_mt
 
    return tmp


#! /usr/bin/env python3
# _*_ coding: utf-8 _*_

#@Author:  LiuXing liuxing2@genomics.cn
#@Date:  2023-07-13 14:30:56
#@Last Modified by: LiuXing
#@Last Modified time: 2023-07-13 14:30:56

import os, sys
import numpy as np
import pandas as pd
import scanpy as sc
import anndata
import seaborn as sns
import matplotlib.pyplot as plt

def umap(adata: anndata,
         size: int = 10,
         color: str = 'cell_neighbor',
         legend_loc: str = 'on data',
         legend_fontsize: int = 20,
         figsize: tuple = (6, 6)
         ):
    if 'umap' not in adata.uns['CN']:
        with plt.rc_context({'axes.facecolor':'white','figure.figsize': figsize}):
            sc.pl.umap(adata, 
                       color=[color],
                       size=size, 
                       ncols = 2, 
                       legend_loc=legend_loc, 
                       legend_fontsize=legend_fontsize)
    else:
        if f"{color}_colors" in adata.uns.keys():
            palette = adata.uns[f'{color}_colors']
        else:
            palette = 'tab20'
        umap_result = adata.uns['CN']['umap']
        leiden_cluster = adata.uns['CN']['leiden_cluster']
        fig, ax = plt.subplots(figsize=figsize)
        sns.scatterplot(x = umap_result[:, 0], y = umap_result[:, 1], hue = leiden_cluster, palette = palette, s = 10, ax=ax)
        ax.set_xlabel('UMAP1')
        ax.set_ylabel('UMAP2')
        ax.set_title('Cellular_Neighborhood')
        plt.show()

def spatial(adata: anndata,
            spot_size: int = 100,
            figsize: list = [6, 6]
            ):
    with plt.rc_context({'axes.facecolor':'white','figure.figsize': figsize}):
            sc.pl.spatial(adata, 
                          color=['cell_neighbor'],
                          size=1.3, 
                          img_key='hires', 
                          alpha=1, 
                          spot_size=spot_size)

def heatmap(adata: anndata,
            cmap: str = 'RdYlBu_r',
            figsize: tuple = (5, 3.5),
            row_cluster: bool = False,
            col_cluster: bool = False,
            z_score = None,
            ):
     cn_pct = adata.uns['CN']['cell_composition']
     sns.clustermap(cn_pct,
                    cmap=cmap,
                    figsize = figsize,
                    row_cluster=row_cluster,
                    col_cluster=col_cluster,
                    z_score=z_score)


#!/usr/bin/env python3
# _*_ coding: utf-8 _*_

# @Author: LiuXing liuxing2@genomics.cn 
# @Date: 2022-06-15 17:27:47 
# @Last Modified by:   LiuXing 
# @Last Modified time: 2022-06-15 17:27:47 

from email.policy import default
import sys, os

import scanpy as sc
import anndata
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

import cell2location
import scvi
from scvi import REGISTRY_KEYS
import matplotlib
from matplotlib import rcParams
rcParams['pdf.fonttype'] = 42 #enables correct plotting of text
import seaborn as sns
from optparse import OptionParser

class Cell2location():
    """
    Run cell type deconvolution or annotation by cell2location, including reference and ST(Spatial Transcriptomics) data procession.

    Parameters
    ----------
    ref_file
        reference file path, which can be csv or anndata format.
    adata_vis_file
        ST file path, which will be deconvoluted or annotated
    out_dir
        output directory path that will be used to save result. Three subdirectory will be created:
        reference_signatures: save processed single-cell reference in csv format if the ref_file was given in anndata format
        cell2location_map: save deconvolution model and result in anndata format
        figures: save figures generated in the deconvolution or annotation process.
    bin_size
        the bin size of ST data in h5ad file, default = 50. If the given data was in single-cell resolution, please specify bin_size=1.
        the N_cells_per_location will be caculated based on this parameter. N_cell_per_location = (bin_size*500/10000)^2 if bins_size > 1 else 1
    gpu
        Load model on default GPU if available (if None or True), or index of GPU to use (if int), or name of GPU (if str), or use CPU (if False).
    """
    def __init__(self, ref_file: str, 
                 adata_vis_file: str,
                 out_dir: str = os.getcwd(),
                 bin_size: int = 50, 
                 gpu = True):
        self.ref_file = ref_file
        self.adata_vis_file = adata_vis_file
        self.results_folder = out_dir
        self.N_cells_per_location = int((bin_size*500/10000)**2) if bin_size > 1 else 1 #calculate the cell number per bin based on bin size
        if str(gpu).upper() == 'TRUE':
            self.gpu = True
        elif str(gpu).upper() == 'FALSE':
            self.gpu = False
        else:
            self.gpu = f"cuda:{gpu}"
        self.ref_run_name = f'{self.results_folder}/reference_signatures'
        self.run_name = f'{self.results_folder}/cell2location_map'
        self.figures = f'{self.results_folder}/figures'
        os.makedirs(self.ref_run_name, exist_ok=True)
        os.makedirs(self.run_name, exist_ok=True)
        os.makedirs(self.figures, exist_ok=True)
        sc.settings.figdir = self.figures

    def run_deconvolution(self):
        """
        run both reference processing and deconvolution with default parameters
        """
        if self.ref_file.endswith(".h5ad"):
            inf_aver = self.process_ref()
        elif self.ref_file.endswith(".csv"):
            inf_aver = pd.read_csv(self.ref_file, index_col=0)
        adata_vis  = self.process_vis(inf_aver)
        return adata_vis

    def process_ref(self,
                    batch_key: str = 'sample',
                    labels_key: str = 'cell_type',
                    max_epochs: int = 1500) -> pd.DataFrame:
        """
        process single-cell sequence reference data

        Parameters
        ----------
        batch_key
            specify the key for obtaining batch infomation. For example, if the reference data collected from different sample, reaction or exprement batch,
            data source should be markered with batch key, and the program will revise the batch effect.
            default = sample
        labels_key
            key name for getting cell type annotation information, default = cell_type.
        max_epochs
            maximal epochs for model training, default=1500.
        
        Returns
        ----------
            infered average gene expression vector of every cell type
        """
        sc.settings.figdir = self.figures
        inf_aver_file = f"{self.ref_run_name}/inf_aver.csv"
        if (os.path.exists(inf_aver_file)):
            inf_aver = pd.read_csv(inf_aver_file, index_col=0)
            return inf_aver
        elif (os.path.exists(f"{self.ref_run_name}/model.pt")):
            adata_ref_file = f"{self.ref_run_name}/sc.h5ad"
            adata_ref = sc.read_h5ad(adata_ref_file)
            mod = cell2location.models.RegressionModel.load(self.ref_run_name, adata_ref, use_gpu=self.gpu)
        else:
            adata_ref = sc.read_h5ad(self.ref_file)
            adata_ref.obs_names_make_unique()
            sc.pp.filter_cells(adata_ref, min_genes = 0)
            sc.pp.filter_genes(adata_ref, min_counts = 10)
            adata_ref.var['mt'] = adata_ref.var_names.str.startswith(('mt-', 'MT-'))
            sc.pp.calculate_qc_metrics(adata_ref, qc_vars=['mt'], percent_top=None, log1p=False, inplace=True)
            sc.pl.violin(adata_ref, ['n_genes', 'total_counts', 'pct_counts_mt'], jitter=0.4, multi_panel=True, save='ref_violin.png')
            sc.pl.scatter(adata_ref, x='total_counts', y='n_genes', save = 'ref_scatter.png')

            from cell2location.utils.filtering import filter_genes
            selected = filter_genes(adata_ref, cell_count_cutoff=5, cell_percentage_cutoff2=0.03, nonz_mean_cutoff=1.12)
            plt.savefig(f"{self.figures}/ref_filter_genes.png")
            plt.clf()
            #filter the object
            adata_ref = adata_ref[:, selected].copy()
            #prepare anndata for the regression model
            cell2location.models.RegressionModel.setup_anndata(adata=adata_ref,
                                                   # reaction / sample / batch
                                                   batch_key = batch_key,
                                                   #batch_key = 'Sample',
                                                   # cell type, covariate used fir contructing signatures
                                                   labels_key = labels_key,
                                                   # multiplicative technical effects (platform, 3' vs 5', donor effect)
                                                   #categorical_covariate_keys = ['Experiment']
                                                   #categorical_covariate_keys = ['Method']
                                                  )
            # create the regression model
            from cell2location.models import RegressionModel
            mod = RegressionModel(adata_ref)
            #vies anndata_setup as a sanity check
            mod.view_anndata_setup()
            mod.train(max_epochs=max_epochs, use_gpu=self.gpu)

            # In this section, we export the estimated cell abundance (summary of the posterior distribution).
            adata_ref = mod.export_posterior(
                adata_ref, sample_kwargs={'num_samples':1000, 'batch_size': 2500, 'use_gpu': self.gpu}
            )

            #Save model
            mod.save(f"{self.ref_run_name}", overwrite=True)

            #Save anndata object with results
            adata_file = f"{self.ref_run_name}/sc.h5ad"
            adata_ref.write(adata_file)
            mod.plot_history(20)
            plt.savefig(f"{self.figures}/ref_train_history.png")
            plt.clf()
            inf_aver = mod.samples[f"post_sample_means"]["per_cluster_mu_fg"].T
            if "detection_y_c" in list(mod.samples[f"post_sample_means"].keys()):
                inf_aver = inf_aver * mod.samples[f"post_sample_means"]["detection_y_c"].mean()
            aver = mod._compute_cluster_averages(key=REGISTRY_KEYS.LABELS_KEY)
            aver = aver[mod.factor_names_]
            plt.hist2d(np.log10(aver.values.flatten() + 1),
                        np.log10(inf_aver.flatten() + 1),
                        bins = 50,
                        norm = matplotlib.colors.LogNorm(),
            )
            plt.xlabel("Mean expression for every gene in every cluster")
            plt.ylabel("Estimated expression for every gene in every cluster")
            plt.savefig(f"{self.figures}/ref_train_QC.png")
            plt.clf()
            #export estimated expression in each cluster
        if 'means_per_cluster_mu_fg' in adata_ref.varm.keys():
            inf_aver = adata_ref.varm['means_per_cluster_mu_fg'][[f'means_per_cluster_mu_fg_{i}'
                                                          for i in adata_ref.uns['mod']['factor_names']]].copy()
        else:
            inf_aver = adata_ref.var[[f'means_per_cluster_mu_fg_{i}'
                                        for i in adata_ref.uns['mod']['factor_names']]].copy()
        inf_aver.columns = adata_ref.uns['mod']['factor_names']
        inf_aver.to_csv(inf_aver_file)
        return inf_aver

    def process_vis(self, 
                    inf_aver: pd.DataFrame, 
                    max_epochs: int = 5000,
                    batch_size: int = 90000,
                    anno: str = 'cell2loc_anno',
                    spot_size: int = 70) -> anndata:
        """
        deconvolute ST data based on the processed SC reference

        Parameters
        ----------
        inf_aver
            infered average gene expression vector of every cell type
        max_epochs
            maximal epochs for model training, default=5000.
        batch_size
            batch size that determines the amount of data loaded into the gpu memory, default=90000. 
            If the running report outOfMemory error, reduce the batch_size can help to resolve.
        anno
            label key that will be used to store annotation result, default=cell2loc_anno.
        spot_size
            specify spot size when draw bins or cells in space.
        Returns
        -----------
            anndata with deconvolution and annotation result.    
        """
        sc.settings.figdir = self.figures
        # find shared genes and subset both anndata and reference signatures
        adata_vis = anndata.read(self.adata_vis_file)
        if adata_vis.raw != None:
            adata_vis = adata_vis.raw.to_adata()
        adata_vis.raw = adata_vis
        #find mitochondria-encoded (MT) genes
        adata_vis.var['MT_gene'] = adata_vis.var_names.str.startswith(('mt-', 'MT-'))

        # remove MT genes for spatial mapping (keeping their counts in the object)
        adata_vis.obsm['MT'] = adata_vis[:, adata_vis.var['MT_gene'].values].X.toarray()
        adata_vis = adata_vis[:, ~adata_vis.var['MT_gene'].values]
        intersect = np.intersect1d(adata_vis.var_names, inf_aver.index)
        adata_vis = adata_vis[:, intersect].copy()
        inf_aver = inf_aver.loc[intersect, :].copy()

        #prepare anndata for cell2location model
        cell2location.models.Cell2location.setup_anndata(adata=adata_vis)

        # create and train the model
        mod = cell2location.models.Cell2location(
            adata_vis, cell_state_df=inf_aver,
            # the expected average cell abundance: tissue-dependent
            # hyper-prior which can be estimated from paired histology:
            N_cells_per_location=self.N_cells_per_location,
            # hyperparameter controlling normalisation of
            # within-experiment variation in RNA detection:
            detection_alpha=20
        )
        mod.view_anndata_setup()
        
        if mod.adata.n_obs < batch_size:
            batch_size = None 
        mod.train(max_epochs=max_epochs,
                # train using full data (batch_size=None)
                batch_size=batch_size,
                # use all data points in training because
                # we need to estimate cell abundance at all locations
                train_size=1,
                use_gpu=self.gpu)

        # In this section, we export the estimated cell abundance (summary of the posterior distribution).
        adata_vis = mod.export_posterior(
            adata_vis, sample_kwargs={'num_samples': 1000, 'batch_size': 10000, 'use_gpu': self.gpu}
        )

        # Save model
        mod.save(f"{self.run_name}", overwrite=True)
        mod.plot_history(20)
        plt.legend(labels=['full data training'])
        plt.savefig(f"{self.figures}/vis_train_history.png")
        plt.clf()

        use_n_obs = 1000
        ind_x = np.random.choice(mod.adata_manager.adata.n_obs, np.min((use_n_obs, mod.adata.n_obs)), replace=False)
        mod.expected_nb_param = mod.module.model.compute_expected(
            mod.samples[f"post_sample_means"], mod.adata_manager, ind_x=ind_x
        )
        x_data = mod.adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY)[ind_x, :]
        x_data = np.asarray(x_data.toarray())
        mod.plot_posterior_mu_vs_data(mod.expected_nb_param["mu"], x_data)
        plt.savefig(f"{self.figures}/vis_QC.png")
        plt.clf()

        # Save anndata object with results
        adata_file = f"{self.run_name}/sp.h5ad"
        #adata_vis.write(adata_file)
        # add 5% quantile, representing confident cell abundance, 'at least this amount is present',
        # to adata.obs with nice names for plotting
        adata_vis.obs[adata_vis.uns['mod']['factor_names']] = adata_vis.obsm['q05_cell_abundance_w_sf']
        cellList=list(set(inf_aver.columns) & set(adata_vis.obs.columns))
        with mpl.rc_context({'axes.facecolor':  'black', 'figure.figsize': [5, 5]}):
            sc.pl.spatial(adata_vis, img_key="hires", color=cellList, spot_size=spot_size, vmin=0, vmax='p99.2', 
                        cmap='magma', save="cell_abundance.png")

        adata_vis.obs[anno] = adata_vis.obs[cellList].idxmax(axis=1)
        
        # Save anndata object with results
        self._plotCells(anno, adata_vis, spot_size=spot_size)
        #self._plotMarkerGenes(anno, adata_vis)
        adata_vis.write(adata_file)
        return adata_vis

    def _plotCells(self,
                   anno: str,
                   adata_vis: anndata,
                   spot_size: int = 70):
        """
        draw the annotation result
        """
        import math
        from tqdm import tqdm
        rcParams['figure.figsize'] = 5, 5
        sc.pl.spatial(adata_vis, img_key="hires", color=anno, spot_size=spot_size, save=f"{anno}.png")

        cellsCount = adata_vis.obs[anno].value_counts()
        cellsRate = cellsCount/cellsCount.sum()*100
        cellsdf = pd.concat([cellsCount, cellsRate], axis=1)
        cellsdf.columns = ['count', 'rate']
        cellsdf.to_csv(f"{self.run_name}/{anno}_cell_count.tsv", sep="\t")
        plt.figure(figsize=(8, 8))
        cellsdf['count'].plot(kind = 'barh')
        i = 0
        sum = cellsdf['count'].sum()
        for _, v in cellsdf['rate'].items():
            plt.text(sum*(v+2)/100, i, '%.2f' % v, ha='center', va='bottom', fontsize=11)
            i+=1
        plt.savefig(f"{self.figures}/{anno}_cell_count.png")
        plt.clf()
        plotcol = 4
        cells = adata_vis.obs[anno].unique()
        plotrow = math.ceil(len(cells)/plotcol)
        figSize = (16, plotrow*4)
        fig = plt.figure(figsize=figSize,dpi=100)
        for j in tqdm(range(len(cells))):
            cell = cells[j]
            i = j+1
            row = int(i/plotcol)
            col = i - row*plotcol
            ax = plt.subplot(plotrow, plotcol, i)
            sc.pl.spatial(adata_vis, img_key="hires", color=anno, groups = [cell], spot_size=spot_size, show=False, ax = ax, title ="{0} ({1})".format(cell, cellsCount.loc[cell]), legend_loc=None)
            ax.set_xlabel("")
            ax.set_ylabel("")
        plt.savefig(f"{self.figures}/{anno}_split.png")
        plt.clf()

def main():
    """
    This program can be used to process scRNAseq reference and annotate ST(spatial transcriptomics) data
    %prog [options]
    """
    parser = OptionParser(main.__doc__)
    parser.add_option("-r", "--reference", action = "store", type = "str", dest = "reference", help = "reference file path, can be h5ad or csv format file.")
    parser.add_option("-i", "--vis", action = "store", type = "str", dest = "vis", help = "input stereo-seq data in h5ad format.")
    parser.add_option("-o", "--outDir", action = "store", type = "str", dest = "outDir", help = "output directory path.")
    parser.add_option("-g", "--gpu", action = "store", type = "int", default = 1, dest = "gpu", help = "give cuda gpu name that will be used to run this program. default=1")
    parser.add_option("--bin_size", action="store", type = int, default = 50, dest = "bin_size", help = "bin size of the given ST data. default=50")

    opts, args = parser.parse_args()

    if (opts.reference == None or opts.vis == None or opts.outDir == None):
        sys.exit(not parser.print_help())

    cell2location = Cell2location(opts.reference, opts.vis, opts.outDir, bin_size=opts.bin_size, gpu = opts.gpu)
    cell2location.run_deconvolution()

if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# _*_ coding: utf-8 _*_

#@Author:  LiuXing liuxing2@genomics.cn
#@Date:  2023-07-12 15:29:29
#@Last Modified by: LiuXing
#@Last Modified time: 2023-07-12 15:29:29

import os, sys
import anndata
import scanpy as sc
import squidpy as sq
import math
import matplotlib.pyplot as plt
from scipy import sparse
import numpy as np
import pandas as pd
from tqdm.notebook import tqdm

def _intensity_show(LRadata: anndata,
                    cells: list,
                    genes: list,
                    l: np.ndarray,
                    r: np.ndarray,
                    key: str = 'expression_spatial_distances',
                    anno = "cell2loc_anno",
                    alpha_g:float = 0.5,
                    alpha_i:float = 0.4,
                    spot_size:float = 2,
                    figsize:tuple = (4, 5),
                    save: str = None):
    """
    draw the spatial cell interaction intensity in space.
    """
    df = pd.DataFrame()
    df['x'] = LRadata.obsm["spatial"][:,0].astype(int)
    df['y'] = LRadata.obsm["spatial"][:,1].astype(int)
    df['intensity'] = LRadata.obsp[key].sum(axis=1).A1*(LRadata.obs[anno].isin(cells).any())
    df[genes[0]] = l
    df[genes[1]] = r
    colors = ['tab:blue', 'tab:orange', 'tab:red']
    fig, ax = plt.subplots(figsize=figsize, dpi=200)
    i=0
    scatters = []
    labels = []
    for gene in genes:
        df1 = df[df[gene]>0]
        scatter = ax.scatter(x =  df1['x'], y = df1['y'], c = colors[i], edgecolors='none', alpha = alpha_g, s = spot_size)
        scatters.append(scatter)
        labels.append(gene)
        i+=1
    if (df.shape[0] > 0):
        scatter = ax.scatter(x = df['x'], y = df['y'], c = colors[2], s = df['intensity']*spot_size, edgecolors='none', alpha = alpha_i)
        scatters.append(scatter)
        labels.append('intensity')
        handles, labels1 = scatter.legend_elements(prop="sizes", alpha=0.6)
        legend1 = ax.legend(handles, labels1, bbox_to_anchor=[1.25, 0.7], fontsize=3, title = 'intensity')
        ax.add_artist(legend1)
    else:
        return 0
    ax.set_ylim(bottom = df['y'].max(), top = 0)
    ax.legend(scatters, labels, prop = {'size': 4}, bbox_to_anchor=[1.25, 0.8], fontsize=4)
    plt.gca().set_aspect(1)
    plt.axis('off')
    plt.title(f"{cells[0]} | {cells[1]}\n({genes[0]} | {genes[1]})")
    if save!=None:
        plt.savefig(save)
    
def intensity_insitu(adata: anndata,
                  cells: list,
                  genes: list,
                  anno: str= 'cell2loc_anno',
                  radius: float = 0.0,
                  distance_coefficient:float = 0.0,
                  connectivities_key: str = "spatial_distances",
                  complex_process_model: str = 'mean',
                  alpha_g:float = 0.5,
                  alpha_i:float = 0.4,
                  spot_size:float = 2,
                  figsize:tuple = (4, 5),
                  save: str = None
                  ) -> int:
    """
    Calculate the spatial cell interaction intensity between specified cells and ligand receptor genes.

    Parameters
    ----------
    adata
        anndata
    cells
        list contains sender and receiver cell type. [sender_cell, receiver_cell]
    genes
        list contains ligand and receptor genes. [ligand, receptor]
    anno
        annotation key, default=cell2loc_anno.
    radius
        radius threshold when constructing nearest neighbor graph. Only be used when the neighbor graph doesn't exist.
    distance_coefficient
        Consider the distance as one of the factor that influence the interaction intensity using the exponential decay formular: C=C0*e^(-k*d).
        The parameter defines the k value in the formular. Default=0, means distance would not influence the interaction intensity.
    connectivities_key
        obtain the constructed nearest neighbor graph in adata.obsp[connectivities_key]. If this doesn't exist, construct
        the neighbor graph with specified radius
    complex_process_model
        determine how to deal with the complexed ligand and receptor which contain multiple subunits. There are two options: mean, min.
        mean: calculate the mean expression of all subunits to represent the complex
        min: pick the minimal expression of all subunits to represent the complex
    
    Returns
    ----------
        intensity value of the entire slide
    """ 
    exp_key = "expression"
    if "_" in genes[0]:
        if complex_process_model == 'min':
            l = (adata.obs[anno]==cells[0])*(adata[:,genes[0].split("_")].X.min(axis=1).toarray()[:,0])
        elif complex_process_model == 'mean':
            l = (adata.obs[anno]==cells[0])*(adata[:,genes[0].split("_")].X.mean(axis=1).A1)
        else:
            raise Exception(f"complex_process_model should be mean or min, but got {complex_process_model}.")
    else:
        l = (adata.obs[anno]==cells[0])*(adata[:,genes[0]].X.sum(axis=1).A1)
    if "_" in genes[1]:
        if complex_process_model == 'min':
            r = (adata.obs[anno]==cells[1])*(adata[:,genes[1].split("_")].X.min(axis=1).toarray()[:,0])
        elif complex_process_model == 'mean':
            r = (adata.obs['cell2loc_anno']==cells[1])*(adata[:,genes[1].split("_")].X.mean(axis=1).A1)
        else:
            raise Exception(f"complex_process_model should be mean or min, but got {complex_process_model}.")
    else:
        r = (adata.obs[anno]==cells[1])*(adata[:,genes[1]].X.sum(axis=1).A1)
    
    if connectivities_key in adata.obsp.keys() and radius == 0:
        connect_matrix = adata.obsp[connectivities_key]
    elif radius > 10:
        key_added = f"{radius}um"
        if f"{key_added}_distances" in adata.obsp.keys():
            connect_matrix = adata.obsp[f"{key_added}_distances"]
        else:
            sq.gr.spatial_neighbors(adata, radius=radius*2, coord_type="generic", key_added=key_added)
            connect_matrix = adata.obsp[f"{key_added}_distances"]
        connectivities_key = f"{key_added}_distances"
    else:
        raise Exception(f"The distances_key ({connectivities_key}) dosn't exist in adata.obsp, and radius has not be specified with a value >= 10")

    l = l.values
    r = r.values
    l_rows = np.where(l > 0)[0]
    r_cols = np.where(r > 0)[0]
    sub_connect_matrix=connect_matrix[l_rows,:][:,r_cols].todense()
    dst = np.where(sub_connect_matrix>0)
    distances = sub_connect_matrix[dst]
    exps = l[l_rows[dst[0]]]*[math.exp(-distance_coefficient*d) for d in distances.A1] + r[r_cols[dst[1]]]
    
    spatial_exp = sparse.csr_matrix((exps, (l_rows[dst[0]], r_cols[dst[1]])), shape=connect_matrix.shape, dtype=int)
    exp_connectivities_key = f"{exp_key}_{connectivities_key}"
    adata.obsp[exp_connectivities_key] = spatial_exp
    neighbors_key = exp_connectivities_key.replace("distances", "neighbors")
    params = adata.uns[connectivities_key.replace("distances", "neighbors")]
    params['weight'] = 'expression'
    adata.uns[neighbors_key] = {'connectivities_key': connectivities_key.replace("distances", "connectivities"),
                                'distances_key': exp_connectivities_key,
                                'params': params,
                                }
    
    _intensity_show(adata, cells, genes, l, r, key=exp_connectivities_key, anno=anno, alpha_g=alpha_g, alpha_i=alpha_i, spot_size=spot_size, figsize=figsize, save=save)
    return spatial_exp.sum()

def intensities_with_radius(adata, pairs = None):
    """
    draw the line plot with radius as x and intensity per area as y

    Parameters
    ---------------
    adata
        anndata
    pairs
        pair list that will be drawed in the plot
        for example Teff|Microphage(APP|CD74)
    """
    try:
        plot_df = adata.uns['intensities_with_radius']
    except Exception:
        print("there must be intensities_with_radius in adata.uns, please run intensity.intensities_with_radius before draw this plot.")
    if pairs == None:
        columns = plot_df.columns.values.tolist()
        columns.remove('radius')
    else:
        columns = pairs

    import matplotlib.colors as mcolors
    colors = list(mcolors.TABLEAU_COLORS.keys())
    i = 0
    for column in columns:
        x = plot_df['radius']
        y = plot_df[column]/(3.14*(plot_df['radius']**2))
        plt.plot(x, y,
                 linestyle="-", linewidth=2, color=colors[i],
                 marker='o', markersize=4, markeredgecolor='black', markerfacecolor=colors[i],
                 label=column,
                 )
        i+=1
    plt.xlabel('radius (µm)')
    plt.ylabel('intensity / area')
    plt.legend()




#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
@File        :   de_gene.py
@Description :   de_gene.py
@Author      :   Liuchuandong liuchuandong@genomics.cn
@Date        :   2023/08/04 17:08:24
"""
import decoupler as dc
import scanpy as sc
import anndata
import sys, os
import numpy as np
import pandas as pd
from pydeseq2.dds import DeseqDataSet
from pydeseq2.ds import DeseqStats

# Only needed for visualization:
import matplotlib.pyplot as plt
import seaborn as sns

def deseq2(adata,sample_col='sample',groups_col='cell_neighbor',
           contrast=['5','Others'],batch=None,save=None):
    '''
    Perform differential gene expression analysis in pseudo-bulk manner by using 'decoupler' and 'pydeseq2' modules.
    Here just give a one line method in StereoSite workflow.

    Parameters
    ----------
    adata
        Anndata objects, make sure .X is raw counts.
    sample_col
        Column of obs contains samples names.
    groups_col
        Column of obs contains groups names.
    contrast
        list contains group names you want to compare,['5','Others'] means CN5 vs Others_CN, while ['5','1'] means CN5 vs CN1
    batch
        Column of obs contains batch names 
    save
        Path to where to save the volcano plot. Infer the filetype if ending on {`.pdf`, `.png`, `.svg`}.
    ----------
    
    Returns
        dataframe of deseq2 result
    
    '''
    # get pseudobulk
    pdata = dc.get_pseudobulk(
        adata,
        sample_col=sample_col,
        groups_col=groups_col,
        mode='sum',
        min_cells=0,
        min_counts=0)
    # Filter genes
    genes = dc.filter_by_expr(pdata, group=groups_col, min_count=0, min_total_count=10)
    pdata = pdata[:, genes].copy()
    if 'Others' in contrast:
        pdata.obs['deGroup'] = [contrast[0] if i==contrast[0] else 'Others' for i in pdata.obs[groups_col]]
    else:
        pdata.obs['deGroup'] = pdata.obs[groups_col]
    # Build DESeq2 object
    if batch is None:
        design_factors = ['deGroup']
    else:
        design_factors=['deGroup',batch]
    dds = DeseqDataSet(
        adata=pdata,
        design_factors=design_factors,
        ref_level=['deGroup', contrast[1]],
        refit_cooks=True,
        n_cpus=8,
    )
    # Compute LFCs
    dds.deseq2()
    # Extract contrast between CN5 vs Others
    stat_res = DeseqStats(dds, contrast=["deGroup", contrast[0], contrast[1]], n_cpus=8)
    # Compute Wald test
    stat_res.summary()
    # Shrink LFCs
    stat_res.lfc_shrink(coeff='deGroup_'+contrast[0]+'_vs_'+contrast[1])
    # Extract results
    results_df = stat_res.results_df
    dc.plot_volcano_df(results_df, x='log2FoldChange', y='padj', top=20,save=save)
    return(results_df)



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
@File        :   ppi_analysis.py
@Description :   ppi_analysis.py
@Author      :   Liuchuandong liuchuandong@genomics.cn
@Date        :   2023/08/01 14:35:03
"""

from gc import DEBUG_COLLECTABLE
import os, sys
from optparse import OptionParser
import numpy as np
import pandas as pd
import networkx as nx
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import subprocess
import logging

PPI_outdir = 'PPI'
os.makedirs(PPI_outdir, exist_ok=True)
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

def mcl(query_net,inflation: float = 4.0):
    '''
    Run Markov CLustering in query net

    Parameters
    ----------
    query_net
        the ppi net queried from STRING database according to user input
    inflation
        the Markov CLuster algorithm's parparmeter, vary this parameter to obtain clusterings at different levels of granularity. A good set of starting values is 1.4, 2, 4, and 6.
    ----------
    '''
    logging.critical("Make sure the Markov CLuster algorithm (MCL,https://github.com/micans/mcl) be installed properly.")
    
    nx.to_pandas_edgelist(query_net).to_csv(PPI_outdir+"/query_net_edgelist.txt",sep='\t',header=0,index=0)
    mcl_result_filepath = PPI_outdir+'/out.query_net_edgelist.txt.I'+str( int(inflation*10) )
    
    subprocess.run(['mcl', PPI_outdir+'/query_net_edgelist.txt','--abc','-I', str(inflation),'-o',mcl_result_filepath], capture_output=True)
    logging.info('mcl result be stored in: '+mcl_result_filepath )
    mcl_result = pd.read_csv(mcl_result_filepath, sep='\t', header=None, index_col=False)

    logging.info('Get {} clusters by markov clustering'.format( len(mcl_result) ))
    logging.info('The lagrest cluster with {} nodes'.format( len(mcl_result.loc[0,].dropna()) ))

    return(mcl_result)



def get_cluster_net(query_net,mcl_result,mcl_id=0):
    '''
    Subset specific net from qurey net according to specific mcl_cluster_id, default is the largest one 
    '''
    cluster_net = nx.subgraph(query_net,[i for i in list(mcl_result.loc[mcl_id,]) if str(i)!='nan'])
    return(cluster_net)



def get_MCC_hub_genes(cluster_net):
    '''
    Find hub genes from specific MCL cluster
    '''
    # 1. get genes by sorting MCC score
    node_MCC  = nx.node_clique_number(cluster_net)
    rank_values = set(node_MCC.values())
    rank_values = sorted(rank_values,reverse=True)
    max_values = rank_values[0:10]
    print('Top10 MCC scores are:',max_values)
    max_MCC_genes = []
    # get genes which with top MCC score, by default output 10 genes, if more than 10 genes with max score, still output
    for i in max_values:
        for k,v in node_MCC.items():
            if v ==i: max_MCC_genes.append(k)
        if len(max_MCC_genes)>10:
            break
    #  2. get genes by sorting nodes degree
    top_degree_genes=[]
    rank_degree = sorted(cluster_net.degree, key=lambda x: x[1], reverse=True)
    degree = list(set([i[1] for i in rank_degree]))
    degree.sort(reverse=True)
    if len(degree)<10:
        degree_cutoff=degree[len(degree)-1]
    else:
        degree_cutoff=degree[9]
    for i in rank_degree:
        if i[1]>=degree_cutoff:
            top_degree_genes.append(i[0])
    print('Top10 degree are:',degree[0:10])
    # 3. get overlapped genes according to MCC and degree
    hub_genes = [i for i in max_MCC_genes if i in top_degree_genes]
    hub_genes_df = pd.DataFrame({'hub_genes':hub_genes})
    rank_degree = dict(rank_degree)
    hub_genes_df['degree'] = [rank_degree[i] for i in hub_genes]
    hub_genes_df['MCC_score'] = [node_MCC[i] for i in hub_genes]
    hub_genes_df = hub_genes_df.sort_values(by='degree',ascending=False,ignore_index=True)
    return(hub_genes_df)



def get_hub_net(hub_genes,cluster_net,cutoff: float = 0.8):
    '''
    Subset hub net of hub gene from specific MCL cluster

    Parpameters
    hub_genes
        The hub genes from specific MCL cluster by using get_hub_gene function.
    cluster_net
        The specific MCL cluster.
    cutoff
        the edge's STRING combined-score cutoff.
    '''
    # 1. remove edges which score<cutoff
    cluster_net=cluster_net.copy()
    for i in list(cluster_net.edges()):
        if cluster_net.edges[i]['combined_score'] < cutoff*1000:
            cluster_net.remove_edge(i[0],i[1])
    # 2. get hub genes and their neighbors
    hub_genes_set=set(hub_genes)
    for i in hub_genes:
        neigb = [n for n in cluster_net.neighbors(i)]
        hub_genes_set = hub_genes_set.union(set(neigb))
    hub_net = nx.subgraph(cluster_net,list(hub_genes_set))
    return(hub_net)




import os, sys
import numpy as np
import pandas as pd
import scanpy as sc
from typing import Any, Mapping, Sequence, TYPE_CHECKING
from matplotlib.axes import Axes
from matplotlib.colorbar import ColorbarBase
import matplotlib.pyplot as plt
from anndata import AnnData
from scanpy import logging as logg, settings
from scipy.sparse import issparse, spmatrix
from scipy.cluster import hierarchy as sch
from pathlib import Path
from typing import Any, Callable, Hashable, Iterable, Sequence, Generator, TYPE_CHECKING
from inspect import signature

class CustomDotplot(sc.pl.DotPlot):

    BASE = 10

    DEFAULT_LARGEST_DOT = 50.0
    DEFAULT_NUM_COLORBAR_TICKS = 5
    DEFAULT_NUM_LEGEND_DOTS = 5

    def __init__(self, minn: float, delta: float, alpha: float, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._delta = delta
        self._minn = minn
        self._alpha = alpha

    def _plot_size_legend(self, size_legend_ax: Axes) -> None:
        y = self.BASE ** -((self.dot_max * self._delta) + self._minn)
        x = self.BASE ** -((self.dot_min * self._delta) + self._minn)
        size_range = -(np.logspace(x, y, self.DEFAULT_NUM_LEGEND_DOTS + 1, base=10).astype(np.float64))
        size_range = (size_range - np.min(size_range)) / (np.max(size_range) - np.min(size_range))
        # no point in showing dot of size 0
        size_range = size_range[1:]

        size = size_range**self.size_exponent
        mult = (self.largest_dot - self.smallest_dot) + self.smallest_dot
        size = size * mult

        # plot size bar
        size_legend_ax.scatter(
            np.arange(len(size)) + 0.5,
            np.repeat(1, len(size)),
            s=size,
            color="black",
            edgecolor="black",
            linewidth=self.dot_edge_lw,
            zorder=100,
        )
        size_legend_ax.set_xticks(np.arange(len(size)) + 0.5)
        labels = [f"{(x * self._delta) + self._minn:.1f}" for x in size_range]
        size_legend_ax.set_xticklabels(labels, fontsize="small")

        # remove y ticks and labels
        size_legend_ax.tick_params(axis="y", left=False, labelleft=False, labelright=False)
        # remove surrounding lines
        for direction in ["right", "top", "left", "bottom"]:
            size_legend_ax.spines[direction].set_visible(False)

        ymax = size_legend_ax.get_ylim()[1]
        size_legend_ax.set_ylim(-1.05 - self.largest_dot * 0.003, 4)
        size_legend_ax.set_title(self.size_title, y=ymax + 0.25, size="small")

        xmin, xmax = size_legend_ax.get_xlim()
        size_legend_ax.set_xlim(xmin - 0.15, xmax + 0.5)

        if self._alpha is not None:
            ax = self.fig.add_subplot()
            ax.scatter(
                [0.35, 0.65],
                [0, 0],
                s=size[-1],
                color="black",
                edgecolor="black",
                linewidth=self.dot_edge_lw,
                zorder=100,
            )
            ax.scatter(
                [0.65], [0], s=0.33 * mult, color="white", edgecolor="black", linewidth=self.dot_edge_lw, zorder=100
            )
            ax.set_xlim([0, 1])
            ax.set_xticks([0.35, 0.65])
            ax.set_xticklabels(["false", "true"])
            ax.set_yticks([])
            ax.set_title(f"significant\n$p={self._alpha}$", y=ymax + 0.25, size="small")
            ax.set(frame_on=False)

            l, b, w, h = size_legend_ax.get_position().bounds
            ax.set_position([l + w, b, w, h])

    def _plot_colorbar(self, color_legend_ax: Axes, normalize: bool) -> None:
        cmap = plt.get_cmap(self.cmap)

        ColorbarBase(
            color_legend_ax,
            orientation="horizontal",
            cmap=cmap,
            norm=normalize,
            ticks=np.linspace(
                np.nanmin(self.dot_color_df.values),
                np.nanmax(self.dot_color_df.values),
                self.DEFAULT_NUM_COLORBAR_TICKS,
            ),
            format="%.2f",
        )

        color_legend_ax.set_title(self.color_legend_title, fontsize="small")
        color_legend_ax.xaxis.set_tick_params(labelsize="small")

def _unique_order_preserving(iterable) -> tuple:
    """Remove items from an iterable while preserving the order."""
    seen: set[Hashable] = set()
    seen_add = seen.add
    return [i for i in iterable if not (i in seen or seen_add(i))], seen

def verbosity(level: int) -> Generator[None, None, None]:
    """
    Temporarily set the verbosity level of :mod:`scanpy`.
    Parameters
    ----------
    level
        The new verbosity level.
    Returns
    -------
    Nothing.
    """
    import scanpy as sc

    verbosity = sc.settings.verbosity
    sc.settings.verbosity = level
    try:
        yield
    finally:
        sc.settings.verbosity = verbosity

def _filter_kwargs(func, kwargs) -> dict:
    style_args = {k for k in signature(func).parameters.keys()}  # noqa: C416
    return {k: v for k, v in kwargs.items() if k in style_args}

def save_fig(fig, path, make_dir= True, ext="png", **kwargs) -> None:
    """
    Save a figure.
    Parameters
    ----------
    fig
        Figure to save.
    path
        Path where to save the figure. If path is relative, save it under :attr:`scanpy.settings.figdir`.
    make_dir
        Whether to try making the directory if it does not exist.
    ext
        Extension to use if none is provided.
    kwargs
        Keyword arguments for :meth:`matplotlib.figure.Figure.savefig`.
    Returns
    -------
    None
        Just saves the plot.
    """
    if os.path.splitext(path)[1] == "":
        path = f"{path}.{ext}"

    path = Path(path)

    if not path.is_absolute():
        path = Path(settings.figdir) / path

    if make_dir:
        try:
            path.parent.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logg.debug(f"Unable to create directory `{path.parent}`. Reason: `{e}`")

    logg.debug(f"Saving figure to `{path!r}`")

    kwargs.setdefault("bbox_inches", "tight")
    kwargs.setdefault("transparent", True)

    fig.savefig(path, **kwargs)

def _dendrogram(data, method, **kwargs):
    link_kwargs = _filter_kwargs(sch.linkage, kwargs)
    dendro_kwargs = _filter_kwargs(sch.dendrogram, kwargs)

    # Row-cluster
    row_link = sch.linkage(data, method=method, **link_kwargs)
    row_dendro = sch.dendrogram(row_link, no_plot=True, **dendro_kwargs)
    row_order = row_dendro["leaves"]

    # Column-cluster
    col_link = sch.linkage(data.T, method=method, **link_kwargs)
    col_dendro = sch.dendrogram(col_link, no_plot=True, **dendro_kwargs)
    col_order = col_dendro["leaves"]

    return row_order, col_order, row_link, col_link


from squidpy._constants._utils import ModeEnum
class DendrogramAxis(ModeEnum):
    INTERACTING_MOLS = "interacting_molecules"
    INTERACTING_CLUSTERS = "interacting_clusters"
    BOTH = "both"

_SEP = " | "

def ligrec(
    adata,
    cluster_key = None,
    source_groups = None,
    target_groups = None,
    intensities_range = (-np.inf, np.inf),
    pvalue_threshold = 1.0,
    remove_empty_interactions = True,
    remove_nonsig_interactions = False,
    dendrogram = None,
    alpha = 0.001,
    swap_axes = False,
    title = None,
    figsize = None,
    dpi = None,
    save = None,
    **kwargs,
) -> None:
    """
    Plot the result of a receptor-ligand permutation test.
    The result was computed by :func:`squidpy.gr.ligrec`.
    :math:`molecule_1` belongs to the source clusters displayed on the top (or on the right, if ``swap_axes = True``,
    whereas :math:`molecule_2` belongs to the target clusters.
    Parameters
    ----------
    %(adata)s
        It can also be a :class:`dict`, as returned by :func:`squidpy.gr.ligrec`.
    %(cluster_key)s
        Only used when ``adata`` is of type :class:`AnnData`.
    source_groups
        Source interaction clusters. If `None`, select all clusters.
    target_groups
        Target interaction clusters. If `None`, select all clusters.
    intensity_range
        Only show interactions whose intensity are within this **closed** interval.
    pvalue_threshold
        Only show interactions with p-value <= ``pvalue_threshold``.
    remove_empty_interactions
        Remove rows and columns that only contain interactions with `NaN` values.
    remove_nonsig_interactions
        Remove rows and columns that only contain interactions that are larger than ``alpha``.
    dendrogram
        How to cluster based on the p-values. Valid options are:
            -  `None` - do not perform clustering.
            - `'interacting_molecules'` - cluster the interacting molecules.
            - `'interacting_clusters'` - cluster the interacting clusters.
            - `'both'` - cluster both rows and columns. Note that in this case, the dendrogram is not shown.
    alpha
        Significance threshold. All elements with p-values <= ``alpha`` will be marked by tori instead of dots.
    swap_axes
        Whether to show the cluster combinations as rows and the interacting pairs as columns.
    title
        Title of the plot.
    %(plotting_save)s
    kwargs
        Keyword arguments for :meth:`scanpy.pl.DotPlot.style` or :meth:`scanpy.pl.DotPlot.legend`.
    Returns
    -------
    %(plotting_returns)s
    """

    def filter_values(
        pvals: pd.DataFrame, intensities: pd.DataFrame, *, mask: pd.DataFrame, kind: str
    ):
        mask_rows = mask.any(axis=1)
        pvals = pvals.loc[mask_rows]
        intensities = intensities.loc[mask_rows]

        if pvals.empty:
            raise ValueError(f"After removing rows with only {kind} interactions, none remain.")

        mask_cols = mask.any(axis=0)
        pvals = pvals.loc[:, mask_cols]
        intensities = intensities.loc[:, mask_cols]

        if pvals.empty:
            raise ValueError(f"After removing columns with only {kind} interactions, none remain.")

        return pvals, intensities

    def get_dendrogram(adata: AnnData, linkage: str = "complete") -> Mapping[str, Any]:
        z_var = sch.linkage(
            adata.X,
            metric="correlation",
            method=linkage,
            optimal_ordering=adata.n_obs <= 1500,  # matplotlib will most likely give up first
        )
        dendro_info = sch.dendrogram(z_var, labels=adata.obs_names.values, no_plot=True)
        # this is what the DotPlot requires
        return {
            "linkage": z_var,
            "groupby": ["groups"],
            "cor_method": "pearson",
            "use_rep": None,
            "linkage_method": linkage,
            "categories_ordered": dendro_info["ivl"],
            "categories_idx_ordered": dendro_info["leaves"],
            "dendrogram_info": dendro_info,
        }

    if dendrogram is not None:
        dendrogram = DendrogramAxis(dendrogram)  # type: ignore[assignment]
        if TYPE_CHECKING:
            assert isinstance(dendrogram, DendrogramAxis)

    if isinstance(adata, AnnData):
        if cluster_key is None:
            raise ValueError("Please provide `cluster_key` when supplying an `AnnData` object.")

        cluster_key = adata.uns.ligrec(cluster_key)
        if cluster_key not in adata.uns_keys():
            raise KeyError(f"Key `{cluster_key}` not found in `adata.uns`.")
        adata = adata.uns[cluster_key]

    if not isinstance(adata, dict):
        raise TypeError(
            f"Expected `adata` to be either of type `anndata.AnnData` or `dict`, found `{type(adata).__name__}`."
        )

    if len(intensities_range) != 2:
        raise ValueError(f"Expected `intensities_range` to be a sequence of size `2`, found `{len(intensities_range)}`.")
    intensities_range = tuple(sorted(intensities_range))  # type: ignore[assignment]

    if alpha is not None and not (0 <= alpha <= 1):
        raise ValueError(f"Expected `alpha` to be in range `[0, 1]`, found `{alpha}`.")

    if source_groups is None:
        source_groups = adata["pvalues"].columns.get_level_values(0)
    elif isinstance(source_groups, str):
        source_groups = (source_groups,)

    if target_groups is None:
        target_groups = adata["pvalues"].columns.get_level_values(1)
    if isinstance(target_groups, str):
        target_groups = (target_groups,)
    if title is None:
        title = "Receptor-ligand test"

    source_groups, _ = _unique_order_preserving(source_groups)  # type: ignore[assignment]
    target_groups, _ = _unique_order_preserving(target_groups)  # type: ignore[assignment]

    pvals: pd.DataFrame = adata["pvalues"].loc[:, (source_groups, target_groups)]
    intensities: pd.DataFrame = adata["intensities"].loc[:, (source_groups, target_groups)]

    if pvals.empty:
        raise ValueError("No valid clusters have been selected.")

    intensities = intensities[(intensities >= intensities_range[0]) & (intensities <= intensities_range[1])]
    pvals = pvals[pvals <= pvalue_threshold]

    if remove_empty_interactions:
        pvals, intensities = filter_values(pvals, intensities, mask=~(pd.isnull(intensities) | pd.isnull(pvals)), kind="NaN")
    if remove_nonsig_interactions and alpha is not None:
        pvals, intensities = filter_values(pvals, intensities, mask=pvals <= alpha, kind="non-significant")

    start, label_ranges = 0, {}

    if dendrogram == DendrogramAxis.INTERACTING_CLUSTERS:
        # rows are now cluster combinations, not interacting pairs
        pvals = pvals.T
        intensities = intensities.T

    for cls, size in (pvals.groupby(level=0, axis=1)).size().to_dict().items():
        label_ranges[cls] = (start, start + size - 1)
        start += size
    label_ranges = {k: label_ranges[k] for k in sorted(label_ranges.keys())}

    pvals = pvals[label_ranges.keys()]
    pvals = -np.log10(pvals + min(1e-3, alpha if alpha is not None else 1e-3)).fillna(0)

    pvals.columns = map(_SEP.join, pvals.columns.to_flat_index())
    pvals.index = map(_SEP.join, pvals.index.to_flat_index())

    intensities = intensities[label_ranges.keys()].fillna(0)
    intensities.columns = map(_SEP.join, intensities.columns.to_flat_index())
    intensities.index = map(_SEP.join, intensities.index.to_flat_index())
    intensities = np.log10(intensities + 1)

    var = pd.DataFrame(pvals.columns)
    var = var.set_index(var.columns[0])

    adata = AnnData(pvals.values, obs={"groups": pd.Categorical(pvals.index)}, var=var, dtype=pvals.values.dtype)
    adata.obs_names = pvals.index
    minn = np.nanmin(adata.X)
    delta = np.nanmax(adata.X) - minn
    adata.X = (adata.X - minn) / delta

    try:
        if dendrogram == DendrogramAxis.BOTH:
            row_order, col_order, _, _ = _dendrogram(
                adata.X, method="complete", metric="correlation", optimal_ordering=adata.n_obs <= 1500
            )
            adata = adata[row_order, :][:, col_order]
            pvals = pvals.iloc[row_order, :].iloc[:, col_order]
            intensities = intensities.iloc[row_order, :].iloc[:, col_order]
        elif dendrogram is not None:
            adata.uns["dendrogram"] = get_dendrogram(adata)
    except IndexError:
        # just in case pandas indexing fails
        raise
    except Exception as e:
        logg.warning(f"Unable to create a dendrogram. Reason: `{e}`")
        dendrogram = None

    kwargs["dot_edge_lw"] = 0
    kwargs.setdefault("cmap", "viridis")
    kwargs.setdefault("grid", True)
    kwargs.pop("color_on", None)  # interferes with tori

    dp = (
        CustomDotplot(
            delta=delta,
            minn=minn,
            alpha=alpha,
            adata=adata,
            var_names=adata.var_names,
            groupby="groups",
            dot_color_df=intensities,
            dot_size_df=pvals,
            title=title,
            var_group_labels=None if dendrogram == DendrogramAxis.BOTH else list(label_ranges.keys()),
            var_group_positions=None if dendrogram == DendrogramAxis.BOTH else list(label_ranges.values()),
            standard_scale=None,
            figsize=figsize,
        )
        .style(
            **_filter_kwargs(sc.pl.DotPlot.style, kwargs),
        )
        .legend(
            size_title=r"$-\log_{10} ~ P$",
            colorbar_title=r"$log_{10}(intensity + 1)$",
            **_filter_kwargs(sc.pl.DotPlot.legend, kwargs),
        )
    )
    if dendrogram in (DendrogramAxis.INTERACTING_MOLS, DendrogramAxis.INTERACTING_CLUSTERS):
        # ignore the warning about mismatching groups
        with verbosity(0):
            dp.add_dendrogram(size=1.6, dendrogram_key="dendrogram")
    if swap_axes:
        dp.swap_axes()

    dp.make_figure()

    if dendrogram != DendrogramAxis.BOTH:
        # remove the target part in: source | target
        labs = dp.ax_dict["mainplot_ax"].get_yticklabels() if swap_axes else dp.ax_dict["mainplot_ax"].get_xticklabels()
        for text in labs:
            text.set_text(text.get_text().split(_SEP)[1])
        if swap_axes:
            dp.ax_dict["mainplot_ax"].set_yticklabels(labs)
        else:
            dp.ax_dict["mainplot_ax"].set_xticklabels(labs)

    if alpha is not None:
        yy, xx = np.where((pvals.values + alpha) >= -np.log10(alpha))
        if len(xx) and len(yy):
            # for dendrogram='both', they are already re-ordered
            mapper = (
                np.argsort(adata.uns["dendrogram"]["categories_idx_ordered"])
                if "dendrogram" in adata.uns
                else np.arange(len(pvals))
            )
            logg.info(f"Found `{len(yy)}` significant interactions at level `{alpha}`")
            ss = 0.33 * (adata.X[yy, xx] * (dp.largest_dot - dp.smallest_dot) + dp.smallest_dot)

            # must be after ss = ..., cc = ...
            yy = np.array([mapper[y] for y in yy])
            if swap_axes:
                xx, yy = yy, xx
            dp.ax_dict["mainplot_ax"].scatter(xx + 0.5, yy + 0.5, color="white", s=ss, lw=0)

    if dpi is not None:
        dp.fig.set_dpi(dpi)

    if save is not None:
        save_fig(dp.fig, save)


import os, sys
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn import datasets
from collections import defaultdict

class PySankeyException(Exception):
    pass


class NullsInFrame(PySankeyException):
    pass


class LabelMismatch(PySankeyException):
    pass

def _check_data_matches_labels(labels, data):
    if isinstance(data, list):
        data = set(data)
    if isinstance(labels, pd.Series):
        data = set(data.unique().tolist())
    if isinstance(labels, list):
        labels = set(labels)
    if labels != data:
        raise LabelMismatch('Labels and data do not match. {0}'.format(','.join(labels)))

def core_process(core:np.array, model_names:list = ['CC_Module', 'LR_Module', 'TME']):
    # preprocess core matrix with three dimensionality: ligand-receptor model, cell-cell model, TME model
    left = []
    mid_l = []
    mid_r = []
    right = []
    left_weight = []
    mid_l_weight = []
    mid_r_weight = []
    right_weight = []
    lr_tme_link = core.sum(axis=0)
    cc_tme_link = core.sum(axis=1)
    lrs, ccs, tmes = core.shape
    for tme in range(tmes):
        for lr in range(lrs):
            left.append(f"{model_names[0]} {lr}")
            mid_l.append(f"{model_names[2]} {tme}")
            left_weight.append(cc_tme_link[lr, tme])
            #mid_l_weight.append(cc_tme_link[lr, tme])
        for cc in range(ccs):
            right.append(f"{model_names[1]} {cc}")
            mid_r.append(f"{model_names[2]} {tme}")
            right_weight.append(lr_tme_link[cc, tme])
            #mid_r_weight.append(lr_tme_link[cc, tme])
    mid_l_weight = tmes*ccs*[cc_tme_link.sum()/(tmes*ccs)]
    mid_r_weight = tmes*lrs*[lr_tme_link.sum()/(tmes*lrs)]
    
    dataFrame = pd.DataFrame({'left': left, 'right': right, 'mid_l': mid_l, 'mid_r': mid_r, 
                          'left_weight': left_weight, 'right_weight': right_weight, 
                              'mid_l_weight': mid_l_weight, 'mid_r_weight': mid_r_weight}, index=range(len(left)))
    return dataFrame

def factor_process(factors:np.array, names:list, model_type:str='CC_Module'):
    left = []
    right = []
    leftWeight = []
    rightWeight = []
    for index, element in np.ndenumerate(factors):
        left.append(f'{model_type} {index[1]}')
        right.append(names[index[0]])
        leftWeight.append(element)
        rightWeight.append(element)
    dataFrame = pd.DataFrame({'left': left, 'right': right, 'leftWeight': leftWeight, 'rightWeight': rightWeight}, index=range(len(left)))
    return dataFrame

def sankey_3d(data:pd.DataFrame, color_dict=None, left_labels = None, right_labels = None, mid_labels = None,
              aspect=3, fontsize=5, save=None, close_plot=False, patch_alpha:float=0.99, link_alpha:float=0.4,
              palette:str='tab20', interval:float=0.005):
    '''
    Make Sankey Diagram showing flow:  ligand-receptor model <--- TME ---> cell-cell model
    
    Inputs:
        data: pandas.dataFrame contains columns left, right, mid_l, mid_r, left_weight, right_weight, mid_l_weight, mid_r_weight
        color_dict: Dictionary of colors to use for each label {'label': 'color'}
        left_labels: Order of the left labels in the diagram
        right_labels: Order of the right labels in the diagram
        mid_labels: Order of the middle labels in the diagram
        aspect: Vertical extent of the diagram in units of horizontal extent
        fontsize: Fontsize of patch label text
        save: If the figure file name was given, the sankey figure will be stored in it.
        palette: Palette style
        interval: Distance between two adjacent patchs = interval * vertical length of all patchs.
        
    Output:
        None
    '''
    plt.figure(dpi=140)
    plt.rc('text', usetex=False)
    plt.rc('font', family='serif')

    if len(data[(data.left.isnull()) | (data.right.isnull()) | (data.mid_l.isnull()) | (data.mid_r.isnull())]):
        raise NullsInFrame('Sankey graph dose not support null values.')

    #Identify all labels that appear 'left' or 'right'
    allLabels = pd.Series(np.r_[data.mid_r.unique(), data.left.unique(), data.right.unique(), data.mid_l.unique()]).unique()

    #Identify labels
    if left_labels == None:
        left_labels = data.left.unique()[::-1]
    else:
        _check_data_matches_labels(left_labels, data['left'])
    if mid_labels == None:
        mid_labels = data.mid_r.unique()[::-1]
    else:
        _check_data_matches_labels(mid_labels, data['mid_l'])
    if right_labels == None:
        right_labels = data.right.unique()[::-1]
    else:
        _check_data_matches_labels(right_labels, data['right'])

    if color_dict is None:
        color_dict = {}
        colorPalette = sns.color_palette(palette, len(allLabels))
        for i, label in enumerate(allLabels):
            color_dict[label] = colorPalette[i]
        #for label in left_labels:
        #    color_dict[label] = colorPalette[i+1]
        #for label in right_labels:
        #    color_dict[label] = colorPalette[i+2]
    else:
        missing = [label for label in allLabels if label not in color_dict.keys()]
        if missing:
            msg = "The color_dict parameter is missing values for the following labels: {}".format(', '.join(missing))
            raise ValueError(msg)

    #Determine widths of individual strips
    
    ns_l = defaultdict()
    ns_m_l = defaultdict()
    ns_m_r = defaultdict()
    ns_r = defaultdict()
    for midLabel in mid_labels:
        leftDict = {}
        midLDict = {}
        midRDict = {}
        rightDict = {}
        for leftLabel in left_labels:
            midLDict[leftLabel] = data[(data.mid_l == midLabel) & (data.left == leftLabel)].mid_l_weight.sum()
            leftDict[leftLabel] = data[(data.mid_l == midLabel) & (data.left == leftLabel)].left_weight.sum()
        ns_m_l[midLabel] = midLDict
        ns_l[midLabel] = leftDict
        for rightLabel in right_labels:
            midRDict[rightLabel] = data[(data.mid_r == midLabel) & (data.right == rightLabel)].mid_r_weight.sum()
            rightDict[rightLabel] = data[(data.mid_r == midLabel) & (data.right == rightLabel)].right_weight.sum()
        ns_m_r[midLabel] = midRDict
        ns_r[midLabel] = rightDict

    midLWidths = defaultdict()
    for i, midLabel in enumerate(mid_labels):
        myD = {}
        myD['mid'] = data[data.mid_l == midLabel].mid_l_weight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['mid']
        else:
            myD['bottom'] = midLWidths[mid_labels[i - 1]]['top'] + interval*data.mid_l_weight.sum()
            myD['top'] = myD['bottom'] + myD['mid']
            topEdge = myD['top']
        midLWidths[midLabel] = myD
    midRWidths = defaultdict()
    for i, midLabel in enumerate(mid_labels):
        myD = {}
        myD['mid'] = data[data.mid_r == midLabel].mid_r_weight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['mid']
        else:
            myD['bottom'] = midRWidths[mid_labels[i - 1]]['top'] + interval*data.mid_r_weight.sum()
            myD['top'] = myD['bottom'] + myD['mid']
            topEdge = myD['top']
        midRWidths[midLabel] = myD

    # Determine positions of left label patches and total widths
    leftWidths = defaultdict()
    for i, leftLabel in enumerate(left_labels):
        myD = {}
        myD['left'] = data[data.left == leftLabel].left_weight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['left']
        else:
            myD['bottom'] = leftWidths[left_labels[i - 1]]['top'] + interval*data.left_weight.sum()
            myD['top'] = myD['bottom'] + myD['left']
            topEdge = myD['top']
        leftWidths[leftLabel] = myD

    # Determine positions of right label patches and total widths
    rightWidths = defaultdict()
    for i, rightLabel in enumerate(right_labels):
        myD = {}
        myD['right'] = data[data.right == rightLabel].right_weight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['right']
        else:
            myD['bottom'] = rightWidths[right_labels[i-1]]['top'] + interval*data.right_weight.sum()
            myD['top'] = myD['bottom'] + myD['right']
            topEdge = myD['top']
        rightWidths[rightLabel] = myD

    #Total vertical extent of diagram
  
    xMax = topEdge/aspect
    width = 0.35*xMax
    pad = 0.9995

    #Draw vertical bars on left and right of each label's section & print label
    for midLabel in mid_labels:
        plt.fill_between(
            [-width/2, width/2],
            2 * [midLWidths[midLabel]['bottom'] *pad],
            2 * [(midLWidths[midLabel]['bottom'] + midLWidths[midLabel]['mid']) * pad],
            color = color_dict[midLabel],
            alpha=patch_alpha
        )
        plt.text(
            0,
            midLWidths[midLabel]['bottom'] + 0.5*midLWidths[midLabel]['mid'],
            midLabel,
            {'ha': 'center', 'va': 'center'},
            fontsize = fontsize
        )

    for leftLabel in left_labels:
        plt.fill_between(
            [-xMax-width, -xMax],
            2 * [leftWidths[leftLabel]['bottom'] * pad],
            2 * [(leftWidths[leftLabel]['bottom'] + leftWidths[leftLabel]['left']) * pad],
            color = color_dict[leftLabel],
            alpha=patch_alpha
        )
        plt.text(
            -xMax - width/2,
            leftWidths[leftLabel]['bottom'] + 0.5*leftWidths[leftLabel]['left'],
            leftLabel,
            {'ha': 'center', 'va': 'center'},
            fontsize=fontsize
        )

    for rightLabel in right_labels:
        plt.fill_between(
            [xMax, xMax + width],
            2 * [rightWidths[rightLabel]['bottom'] * pad],
            2 * [(rightWidths[rightLabel]['bottom'] + rightWidths[rightLabel]['right']) * pad],
            color = color_dict[rightLabel],
            alpha=patch_alpha
        )
        plt.text(
            xMax + width/2,
            rightWidths[rightLabel]['bottom'] + 0.5*rightWidths[rightLabel]['right'],
            rightLabel,
            {'ha': 'center', 'va': 'center'},
            fontsize = fontsize
        )

    # Plot strips
    for midLabel in mid_labels:
        for leftLabel in left_labels:
            labelColor = midLabel
            if len(data[(data.mid_l == midLabel)& (data.left == leftLabel)]) > 0:
                # Create array of y values for each strip, half at middle value, half at left
                ys_d = np.array(50 * [leftWidths[leftLabel]['bottom']] + 50*[midLWidths[midLabel]['bottom']])
                ys_d = np.convolve(ys_d, 0.05*np.ones(20), mode='valid')
                ys_d = np.convolve(ys_d, 0.05*np.ones(20), mode='valid')
                ys_u = np.array(50*[leftWidths[leftLabel]['bottom'] + ns_l[midLabel][leftLabel]] + 50*[midLWidths[midLabel]['bottom'] + ns_m_l[midLabel][leftLabel]])
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')

                # Update bottom edges at each label so next strip starts at the right place
                midLWidths[midLabel]['bottom'] += ns_m_l[midLabel][leftLabel]
                leftWidths[leftLabel]['bottom'] += ns_l[midLabel][leftLabel]
                plt.fill_between(
                    np.linspace(-xMax, -width/2, len(ys_d)), ys_d, ys_u, alpha=link_alpha,
                    color = color_dict[labelColor]
                )

        for rightLabel in right_labels:
            labelColor = midLabel
            if len(data[(data.mid_r == midLabel) & (data.right == rightLabel)]) > 0 :
                # Create array of y values for each strip, half at let value,
                # half at right
                ys_d = np.array(50 * [midRWidths[midLabel]['bottom']] + 50 * [rightWidths[rightLabel]['bottom']])
                ys_d = np.convolve(ys_d, 0.05*np.ones(20), mode='valid')
                ys_d = np.convolve(ys_d, 0.05*np.ones(20), mode='valid')
                ys_u = np.array(50 * [midRWidths[midLabel]['bottom'] + ns_m_r[midLabel][rightLabel]] + 50 * [rightWidths[rightLabel]['bottom'] + ns_r[midLabel][rightLabel]])
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')

                # Update bottom edges at each label so next strip starts at the right place
                midRWidths[midLabel]['bottom'] += ns_m_r[midLabel][rightLabel]
                rightWidths[rightLabel]['bottom'] += ns_r[midLabel][rightLabel]
                plt.fill_between(
                    np.linspace(width/2, xMax, len(ys_d)), ys_d, ys_u, alpha=link_alpha,
                    color = color_dict[labelColor]
                )
    plt.gca().axis('off')
    plt.gcf().set_size_inches(6, 6)
    if save != None:
        plt.savefig(save, bbox_inches='tight', dpi=150)
    if close_plot:
        plt.close()

def sankey_2d(data:pd.DataFrame, color_dict=None, left_labels = None, right_labels = None,
              aspect=4, fontsize=4, save=None, close_plot=False, patch_alpha:float=0.99, link_alpha:float=0.65,
              interval:float=0.0
              ):
    '''
    Make Sankey Diagram
    
    Inputs:
        data: pandas.DataFrame contains columns left, right, mid_l, mid_r, left_weight, right_weight, mid_l_weight, mid_r_weight
        color_dict: Dictionary of colors to use for each label {'label': 'color'}
        left_labels: order of the left labels in the diagram
        right_labels: order of the right labels in the diagram
        aspect: vertical extent of the diagram in units of horizontal extent
        
    Output:
        None
    '''  
    plt.figure(dpi=140)
    plt.rc('text', usetex=False)
    plt.rc('font', family='serif')

    if len(data[(data.left.isnull()) | (data.right.isnull())]):
        raise NullsInFrame('Sankey graph dose not support null values.')

    #Identify all labels that appear 'left' or 'right'
    allLabels = pd.Series(np.r_[data.left.unique(), data.right.unique()]).unique()

    #Identify left labels
    if left_labels == None:
        left_labels = data.left.unique()[::-1]
    else:
        _check_data_matches_labels(left_labels, data['left'])
    if right_labels == None:
        right_labels = data.right.unique()[::-1]
    else:
        _check_data_matches_labels(right_labels, data['right'])

    color_dict = {}
    palette = 'tab20'
    colorPalette = sns.color_palette(palette, len(allLabels))
    for i, label in enumerate(allLabels):
        color_dict[label] = colorPalette[i]

    #Determine widths of individual strips
    from collections import defaultdict
    ns_l = defaultdict()
    ns_r = defaultdict()
    for leftLabel in left_labels:
        leftDict = {}
        rightDict = {}
        for rightLabel in right_labels:
            leftDict[rightLabel] = data[(data.left == leftLabel) & (data.right == rightLabel)].leftWeight.sum()
            rightDict[rightLabel] = data[(data.left == leftLabel) & (data.right == rightLabel)].rightWeight.sum()
        ns_l[leftLabel] = leftDict
        ns_r[leftLabel] = rightDict

    # Determine positions of left label patches and total widths
    leftWidths = defaultdict()
    for i, leftLabel in enumerate(left_labels):
        myD = {}
        myD['left'] = data[data.left == leftLabel].leftWeight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['left']
        else:
            myD['bottom'] = leftWidths[left_labels[i - 1]]['top'] + interval*data.leftWeight.sum()
            myD['top'] = myD['bottom'] + myD['left']
            topEdge = myD['top']
        leftWidths[leftLabel] = myD

    # Determine positions of right label patches and total widths
    rightWidths = defaultdict()
    for i, rightLabel in enumerate(right_labels):
        myD = {}
        myD['right'] = data[data.right == rightLabel].rightWeight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['right']
        else:
            myD['bottom'] = rightWidths[right_labels[i-1]]['top'] + interval*data.rightWeight.sum()
            myD['top'] = myD['bottom'] + myD['right']
            topEdge = myD['top']
        rightWidths[rightLabel] = myD

    #Total vertical extent of diagram
    l_width = -0.3
    r_width = 1.3
    xMax = topEdge/aspect

    #Draw vertical bars on left and right of each label's section & print label
    for leftLabel in left_labels:
        plt.fill_between(
            [l_width*xMax, 0],
            2 * [leftWidths[leftLabel]['bottom']],
            2 * [leftWidths[leftLabel]['bottom'] + leftWidths[leftLabel]['left']],
            color = color_dict[leftLabel],
            alpha=patch_alpha
        )
        plt.text(
            l_width/2 * xMax,
            leftWidths[leftLabel]['bottom'] + 0.5*leftWidths[leftLabel]['left'],
            leftLabel,
            {'ha': 'center', 'va': 'center'},
            fontsize=fontsize
        )
    for rightLabel in right_labels:
        plt.fill_between(
            [xMax, r_width*xMax], 2 * [rightWidths[rightLabel]['bottom']],
            2 * [rightWidths[rightLabel]['bottom'] + rightWidths[rightLabel]['right']],
            color = color_dict[rightLabel],
            alpha=patch_alpha
        )
        plt.text(
            (r_width+l_width/2) * xMax,
            rightWidths[rightLabel]['bottom'] + 0.5*rightWidths[rightLabel]['right'],
            rightLabel,
            {'ha': 'center', 'va': 'center'},
            fontsize = fontsize
        )

    # Plot strips
    for leftLabel in left_labels:
        for rightLabel in right_labels:
            labelColor = leftLabel
            if len(data[(data.left == leftLabel) & (data.right == rightLabel)]) >0 :
                # Create array of y values for each strip, half at let value,
                # half at right
                ys_d = np.array(50 * [leftWidths[leftLabel]['bottom']] + 50 * [rightWidths[rightLabel]['bottom']])
                ys_d = np.convolve(ys_d, 0.05*np.ones(20), mode='valid')
                ys_d = np.convolve(ys_d, 0.05*np.ones(20), mode='valid')
                ys_u = np.array(50 * [leftWidths[leftLabel]['bottom'] + ns_l[leftLabel][rightLabel]] + 50 * [rightWidths[rightLabel]['bottom'] + ns_r[leftLabel][rightLabel]])
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')

                # Update bottom edges at each label so next strip starts at the right place
                leftWidths[leftLabel]['bottom'] += ns_l[leftLabel][rightLabel]
                rightWidths[rightLabel]['bottom'] += ns_r[leftLabel][rightLabel]
                plt.fill_between(
                    np.linspace(0, xMax, len(ys_d)), ys_d, ys_u, alpha=link_alpha,
                    color = color_dict[labelColor]
                )
    plt.gca().axis('off')
    plt.gcf().set_size_inches(6, 6)
    if save != None:
        plt.savefig(save, bbox_inches='tight', dpi=150)
    if close_plot:
        plt.close()


from  setuptools import setup, find_packages

with open("PYPI.rst", "r") as f:
    long_description = f.read()

setup(
    name='stereosite',
    version='2.1.1',
    author='LiuXing',
    author_email='liuxing2@genomics.cn',
    description=('Analysis spatial transcriptomics data'),
    long_description=long_description,
    license='GPL-3 License',
    keywords='spatial cell interaction intensity',
    url="https://github.com/STOmics/StereoSiTE",

    packages=find_packages(), #['stereosite'], #需要打包的目录列表

    include_package_data=True,
    platforms='any',
    #需要安装的依赖包
    install_requires = [
        'anndata>=0.8.0',
        'scanpy>=1.9.1',
        'squidpy>=1.1.2',
        'decoupler>=1.4.0',
        'pydeseq2>=0.3.6',
        'networkx>=3.1',
        'tensorly>=0.8.1',
        'scikit-learn>=1.2.1',
        'torch>=1.11.0',
        'igraph>=0.10.4',
        'pycirclize>=1.1.0',
        'cell2location==0.1.3'
    ],
)



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
@File        :   tf_infer.py
@Description :   tf_infer.py
@Author      :   Liuchuandong liuchuandong@genomics.cn
@Date        :   2023/08/07 10:26:24
"""
import decoupler as dc
import pandas as pd

def ulm(de_result,grn,weight,save):
    '''
    Inferring transcription factor (TF) activities from our transcriptomics data by using decoupler's Univariate Linear Model(ulm) method.
    
    Parameters
    ----------
    de_result
        dataframe of deseq2 result.
    grn
        Gene regulatory network from CollecTRI.
    weight
        Column name in GRN net with weights.
    save
        Path to where to save the plot. Infer the filetype if ending on {`.pdf`, `.png`, `.svg`}.
    ----------

    Returns
        dataframe contains TF enrichment scores and pvalues
    '''
    
    mat = de_result[['stat']].T
    tf_acts, tf_pvals = dc.run_ulm(mat=mat, net=grn, weight=weight,verbose=True)
    tf_df = pd.DataFrame({'tf_acts':tf_acts.T['stat'],
                          'tf_pvals':tf_pvals.T['stat']},index=tf_acts.columns)
    dc.plot_barplot(tf_acts, 'stat', top=25, vertical=True,save=save)

    return(tf_df)


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
@File        :   get_hub_net.py
@Description :   get_hub_net.py
@Author      :   Liuchuandong liuchuandong@genomics.cn
@Date        :   2023/08/01 11:42:43
"""



import os, sys
from optparse import OptionParser
import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

POS_SEED=99
HUB_NODE_SIZE = 366
BASIC_NODE_SIZE = 100
PPI_outdir = 'PPI'
os.makedirs(PPI_outdir, exist_ok=True)
TF_outdir = 'TF'
os.makedirs(TF_outdir, exist_ok=True)

def _edge_colors(graph,cmap='binary',min_weight=0.4,max_weight=1,unit=0.001):
    '''
    set edges color by STRING combined_score
    '''
    norm=matplotlib.colors.Normalize(vmin=min_weight, vmax=max_weight)
    scale_map = matplotlib.cm.ScalarMappable(cmap=cmap, norm=norm)
    colors = []
    for i in graph.edges():
        weight = graph.edges[i]['combined_score']
        colors.append( matplotlib.colors.rgb2hex(scale_map.to_rgba(weight*unit)) )
    return({'color_list':colors,'scale_colormap':scale_map})

def _node_colors(graph,logFC_df,name,cmap='Reds',min_weight=None,max_weight=None,unit=0.1):
    '''
    set node colors by logFC
    logFC_df
        Dataframe contains gene's logFoldChange column,and node names as row index
    name
        Column name in logFC_df with gene's logFoldChange
    '''
    logFC_df = logFC_df.loc[list(graph.nodes()),:]
    for i in graph.nodes():
        graph.nodes[i]['logFC']=logFC_df.loc[i,name]
    if min_weight==None:
        min_weight= round(min(logFC_df[name]),1)
    if max_weight==None:
        max_weight= round(max(logFC_df[name]),1)
        
    norm=matplotlib.colors.Normalize(vmin=min_weight, vmax=max_weight)
    scale_map = matplotlib.cm.ScalarMappable(cmap=cmap, norm=norm)
    colors = []
    for i in graph.nodes():
        fc = round(graph.nodes[i]['logFC'],1)
        colors.append( matplotlib.colors.rgb2hex(scale_map.to_rgba(fc)) )
    return({'color_list':colors,'scale_colormap':scale_map})


def ppi_hub_net(hub_net,hub_gene,logFC_df,name,save=PPI_outdir+'hub_net_cluster1.pdf',figsize=(6,6.8),hspace = 0.4,
                 edge_min_weight=0.8,edge_max_weight=1.0,edge_unit=0.001,edge_cmap='binary',
                 node_min_weight=0,node_max_weight=None,node_unit=0.1,node_cmap='Reds'):
    '''
    plot ppi network

    Parameters
    ----------
    hub_net
        Networkx object, the net need to plot.
    hub_gene
        The hub genes of ppi net.
    logFC_df
        Differential analysis result contains gene's logFoldChange column.
    name
        Column name in logFC_df with gene's logFoldChange
    ----------
    '''
    fig, ax=plt.subplots(3,1,figsize=figsize,height_ratios=[28,1,1])
    edge_colors = _edge_colors(hub_net,cmap=edge_cmap,min_weight=edge_min_weight,max_weight=edge_max_weight,unit=edge_unit)
    node_colors = _node_colors(hub_net,logFC_df=logFC_df,name=name,cmap=node_cmap,min_weight=node_min_weight,max_weight=node_max_weight,unit=node_unit)
    pos = nx.spring_layout(hub_net, seed=POS_SEED)  # Seed layout for reproducibility
    nx.draw(hub_net,pos,edge_color=edge_colors['color_list'],node_color=node_colors['color_list'],ax=ax[0],
            with_labels=True, font_weight='bold',font_size=9,node_size=[HUB_NODE_SIZE if v in hub_gene else BASIC_NODE_SIZE for v in hub_net.nodes()])
    fig.colorbar(edge_colors['scale_colormap'], cax=ax[1], orientation='horizontal', label='STRING combined_score')
    fig.colorbar(node_colors['scale_colormap'], cax=ax[2], orientation='horizontal', label='logFC')
    plt.subplots_adjust(hspace = hspace)
    plt.savefig(save)

def _select_grn(tfs,logFC_df,grn,source='source',target='target',pathway_genes=None):
    '''
    Select nodes from input tfs and input pathway

    Parameters
    ----------
    tfs
        Transcription factor names in GRN.
    grn
        Dataframe of gene regulatory network (GRN).
    source
        Column names in grn dataframe with source nodes.
    target
        Column names in grn dataframe with target nodes.
    logFC_df
        Dataframe of deseq2 result.
    pathway_genes
        Genes list of pathway.
    ----------

    Returns
        Dataframe of selected gene regulatory network (GRN)
    '''
    if pathway_genes==None:
        select_grn=grn.loc[[i for i in grn.index if grn.loc[i,source] in tfs],:].reset_index(drop=True)
        select_grn=select_grn.loc[[i for i in select_grn.index if select_grn.loc[i,target] in logFC_df.index],:].reset_index(drop=True)
    else:
        select_grn=grn.loc[[i for i in grn.index if grn.loc[i,source] in tfs],:].reset_index(drop=True)
        pathway_genes = [i for i in pathway_genes if i in logFC_df.index]
        select_grn=select_grn.loc[[i for i in select_grn.index if select_grn.loc[i,target] in pathway_genes],:].reset_index(drop=True)
    return(select_grn)

def tf_net(tfs,grn,logFC_df,name,source='source',target='target',pathway_genes=None,save=TF_outdir+'/TF_net.pdf',
           figsize=(6.8,6),wspace = 0.04,node_cmap='Reds',node_min_weight=None,node_max_weight=None,node_unit=0.1):
    '''
    Plot gene regulatory network (GRN) of input transcription factors. Node color represent logFC, larger Nodes represent input TFs.

    Parameters
    ----------
    tfs
        Transcription factor names in GRN.
    grn
        Gene regulatory network (GRN).
    logFC_df
        dataframe of deseq2 result.
    name
        Column name in logFC_df with gene's logFoldChange.
    source
        Column names in grn dataframe with source nodes.
    target
        Column names in grn dataframe with target nodes.
    pathway_genes
        Genes list of pathway. Optional.
    save
        Path to where to save the plot. Infer the filetype if ending on {`.pdf`, `.png`, `.svg`}.
    ----------

    '''
    select_grn = _select_grn(tfs=tfs,logFC_df=logFC_df,grn=grn,source=source,target=target,pathway_genes=pathway_genes)
    tf_net = nx.from_pandas_edgelist(df=select_grn,source=source,target=target,create_using=nx.DiGraph())
    node_colors = _node_colors(graph=tf_net,logFC_df=logFC_df,name=name,cmap=node_cmap,min_weight=node_min_weight,max_weight=node_max_weight,unit=node_unit)
    fig, ax=plt.subplots(1,2,figsize=figsize,width_ratios=[28,1])
    pos = nx.circular_layout(tf_net)
    nx.draw(tf_net,pos,ax=ax[0],connectionstyle="arc3,rad=0.1",with_labels=True, edge_color='grey',width=0.5,arrowsize=8,
            #edgecolors=['red' if results_df.loc[v,'padj']<0.05 else node_colors['color_list'][list(tf_net.nodes()).index(v)] for v in tf_net.nodes()],
            font_size=9,font_weight='normal',verticalalignment='baseline',
            node_color=node_colors['color_list'],
            node_size=[HUB_NODE_SIZE if v in tfs else BASIC_NODE_SIZE for v in tf_net.nodes()])

    fig.colorbar(node_colors['scale_colormap'], cax=ax[1], orientation='vertical', label='logFC')#horizontal
    plt.subplots_adjust(wspace = wspace)
    plt.savefig(save)

